from typing import Any, Callable, Dict, List, Optional, Union, Tuple

from deltacat.storage import (
    EntryParams,
    EntryType,
    Delta,
    DeltaLocator,
    DeltaProperties,
    DeltaType,
    DistributedDataset,
    LifecycleState,
    ListResult,
    LocalDataset,
    LocalTable,
    ManifestAuthor,
    Namespace,
    NamespaceProperties,
    Partition,
    PartitionLocator,
    PartitionScheme,
    PartitionValues,
    Schema,
    SortScheme,
    Stream,
    StreamFormat,
    StreamLocator,
    Table,
    TableProperties,
    TableVersion,
    TableVersionLocator,
    TableVersionProperties,
)
from deltacat.storage.model.manifest import Manifest
from deltacat.storage.model.partition import UNKNOWN_PARTITION_ID
from deltacat.types.media import (
    ContentType,
    DistributedDatasetType,
    StorageType,
    DatasetType,
)
from deltacat.utils.common import ReadKwargsProvider


def list_namespaces(*args, **kwargs) -> ListResult[Namespace]:
    """
    Lists a page of table namespaces. Namespaces are returned as list result
    items.
    """
    raise NotImplementedError("list_namespaces not implemented")


def list_tables(namespace: str, *args, **kwargs) -> ListResult[Table]:
    """
    Lists a page of tables for the given table namespace. Tables are returned as
    list result items. Raises an error if the given namespace does not exist.
    """
    raise NotImplementedError("list_tables not implemented")


def list_table_versions(
    namespace: str, table_name: str, *args, **kwargs
) -> ListResult[TableVersion]:
    """
    Lists a page of table versions for the given table. Table versions are
    returned as list result items. Raises an error if the given table does not
    exist.
    """
    raise NotImplementedError("list_table_versions not implemented")


def list_streams(
    namespace: str,
    table_name: str,
    table_version: str,
    *args,
    **kwargs,
) -> ListResult[Stream]:
    """
    Lists a page of streams for the given table version.
    Raises an error if the table version does not exist.
    """
    raise NotImplementedError("list_streams not implemented")


def list_partitions(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> ListResult[Partition]:
    """
    Lists a page of partitions for the given table version. Partitions are
    returned as list result items. Table version resolves to the latest active
    table version if not specified. Raises an error if the table version does
    not exist.
    """
    raise NotImplementedError("list_partitions not implemented")


def list_stream_partitions(stream: Stream, *args, **kwargs) -> ListResult[Partition]:
    """
    Lists all partitions committed to the given stream.
    """
    raise NotImplementedError("list_stream_partitions not implemented")


def list_deltas(
    namespace: str,
    table_name: str,
    partition_values: Optional[PartitionValues] = None,
    table_version: Optional[str] = None,
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: Optional[bool] = None,
    include_manifest: bool = False,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> ListResult[Delta]:
    """
    Lists a page of deltas for the given table version and committed partition.
    Deltas are returned as list result items. Deltas returned can optionally be
    limited to inclusive first and last stream positions. Deltas are returned by
    descending stream position by default. Table version resolves to the latest
    active table version if not specified. Partition values should not be
    specified for unpartitioned tables. Partition scheme ID resolves to the
    table version's current partition scheme by default. Raises an error if the
    given table version or partition does not exist.

    To conserve memory, the deltas returned do not include manifests by
    default. The manifests can either be optionally retrieved as part of this
    call or lazily loaded via subsequent calls to `get_delta_manifest`.
    """
    raise NotImplementedError("list_deltas not implemented")


def list_partition_deltas(
    partition_like: Union[Partition, PartitionLocator],
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: bool = False,
    include_manifest: bool = False,
    *args,
    **kwargs,
) -> ListResult[Delta]:
    """
    Lists a page of deltas committed to the given partition.

    To conserve memory, the deltas returned do not include manifests by
    default. The manifests can either be optionally retrieved as part of this
    call or lazily loaded via subsequent calls to `get_delta_manifest`.
    """
    raise NotImplementedError("list_partition_deltas not implemented")


def get_delta(
    namespace: str,
    table_name: str,
    stream_position: int,
    partition_values: Optional[PartitionValues] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Delta]:
    """
    Gets the delta for the given table version, partition, and stream position.
    Table version resolves to the latest active table version if not specified.
    Partition values should not be specified for unpartitioned tables. Partition
    scheme ID resolves to the table version's current partition scheme by
    default. Raises an error if the given table version or partition does not
    exist.

    To conserve memory, the delta returned does not include a manifest by
    default. The manifest can either be optionally retrieved as part of this
    call or lazily loaded via a subsequent call to `get_delta_manifest`.
    """
    raise NotImplementedError("get_delta not implemented")


def get_latest_delta(
    namespace: str,
    table_name: str,
    partition_values: Optional[PartitionValues] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Delta]:
    """
    Gets the latest delta (i.e. the delta with the greatest stream position) for
    the given table version and partition. Table version resolves to the latest
    active table version if not specified. Partition values should not be
    specified for unpartitioned tables. Partition scheme ID resolves to the
    table version's current partition scheme by default. Raises an error if the
    given table version or partition does not exist.

    To conserve memory, the delta returned does not include a manifest by
    default. The manifest can either be optionally retrieved as part of this
    call or lazily loaded via a subsequent call to `get_delta_manifest`.
    """
    raise NotImplementedError("get_latest_delta not implemented")


def download_delta(
    delta_like: Union[Delta, DeltaLocator],
    table_type: DatasetType = DatasetType.PYARROW,
    storage_type: StorageType = StorageType.DISTRIBUTED,
    max_parallelism: Optional[int] = None,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    distributed_dataset_type: DistributedDatasetType = DistributedDatasetType.RAY_DATASET,
    *args,
    **kwargs,
) -> Union[LocalDataset, DistributedDataset]:  # type: ignore
    """
    Reads the given delta or delta locator into either a list of
    tables resident in the local node's memory, or into a dataset distributed
    across this Ray cluster's object store memory. Ordered table N of a local
    table list, or ordered block N of a distributed dataset, always contain
    the contents of ordered delta manifest entry N.
    """
    raise NotImplementedError("download_delta not implemented")


def download_delta_manifest_entry(
    delta_like: Union[Delta, DeltaLocator],
    entry_index: int,
    table_type: DatasetType = DatasetType.PYARROW,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    *args,
    **kwargs,
) -> LocalTable:
    """
    Reads a single manifest entry into the specified table type for the
    given delta or delta locator. If a delta is provided with a non-empty
    manifest, then the entry is read from this manifest. Otherwise, the
    manifest is first retrieved then the given entry index read.

    NOTE: The entry will be read in the current node's memory.
    """
    raise NotImplementedError("download_delta_manifest_entry not implemented")


def get_delta_manifest(
    delta_like: Union[Delta, DeltaLocator], *args, **kwargs
) -> Manifest:
    """
    Get the manifest associated with the given delta or delta locator. This
    always retrieves the authoritative durable copy of the delta manifest, and
    never the local manifest defined for any input delta. Raises an error if
    the delta can't be found, or if it doesn't contain a manifest.
    """
    raise NotImplementedError("get_delta_manifest not implemented")


def create_namespace(
    namespace: str,
    properties: Optional[NamespaceProperties] = None,
    *args,
    **kwargs,
) -> Namespace:
    """
    Creates a table namespace with the given name and properties. Returns
    the created namespace.
    """
    raise NotImplementedError("create_namespace not implemented")


def update_namespace(
    namespace: str,
    properties: Optional[NamespaceProperties] = None,
    new_namespace: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    """
    Updates a table namespace's name and/or properties. Raises an error if the
    given namespace does not exist.
    """
    raise NotImplementedError("update_namespace not implemented")


def create_table_version(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = LifecycleState.CREATED,
    schema: Optional[Schema] = None,
    partition_scheme: Optional[PartitionScheme] = None,
    sort_keys: Optional[SortScheme] = None,
    table_version_description: Optional[str] = None,
    table_version_properties: Optional[TableVersionProperties] = None,
    table_description: Optional[str] = None,
    table_properties: Optional[TableProperties] = None,
    supported_content_types: Optional[List[ContentType]] = None,
    *args,
    **kwargs,
) -> Tuple[Table, TableVersion, Stream]:
    """
    Create a table version with the given or CREATED lifecycle state and an empty delta
    stream. Table versions may be schemaless and unpartitioned to improve write
    performance, or have their writes governed by a schema and partition scheme
    to improve data consistency and read performance.

    Returns a tuple containing the created/updated table, table version, and
    stream (respectively).

    Raises an error if the given namespace does not exist.
    """
    raise NotImplementedError("create_table_version not implemented")


def create_table(
    namespace: str,
    table_name: str,
    description: Optional[str] = None,
    properties: Optional[TableProperties] = None,
    *args,
    **kwargs,
) -> Table:
    """
    Create a new table. Raises an error if the given table already exists.
    """
    raise NotImplementedError("create_table not implemented")


def update_table(
    namespace: str,
    table_name: str,
    description: Optional[str] = None,
    properties: Optional[TableProperties] = None,
    new_table_name: Optional[str] = None,
    *args,
    **kwargs,
) -> Table:
    """
    Update table metadata describing the table versions it contains. By default,
    a table's properties are empty, and its description is equal to that given
    when its first table version was created. Raises an error if the given
    table does not exist.
    """
    raise NotImplementedError("update_table not implemented")


def update_table_version(
    namespace: str,
    table_name: str,
    table_version: str,
    lifecycle_state: Optional[LifecycleState] = None,
    schema: Optional[Schema] = None,
    description: Optional[str] = None,
    properties: Optional[TableVersionProperties] = None,
    partition_scheme: Optional[PartitionScheme] = None,
    # TODO(pdames): rename to `sort_scheme`
    sort_keys: Optional[SortScheme] = None,
    *args,
    **kwargs,
) -> Tuple[Optional[Table], TableVersion, Optional[Stream]]:
    """
    Update a table version. Notably, updating an unreleased table version's
    lifecycle state to 'active' telegraphs that it is ready for external
    consumption, and causes all calls made to consume/produce streams,
    partitions, or deltas from/to its parent table to automatically resolve to
    this table version by default (i.e. when the client does not explicitly
    specify a different table version). Raises an error if the given table
    version does not exist.
    """
    raise NotImplementedError("update_table_version not implemented")


def stage_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    *args,
    **kwargs,
) -> Stream:
    """
    Stages a new delta stream for the given table version. Resolves to the
    latest active table version if no table version is given. Resolves to the
    DeltaCAT stream format if no stream format is given. If this stream
    will replace another stream with the same format and scheme, then it will
    have its previous stream ID set to the ID of the stream being replaced.
    Returns the staged stream. Raises an error if the table version does not
    exist.
    """
    raise NotImplementedError("stage_stream not implemented")


def commit_stream(
    stream: Stream,
    *args,
    **kwargs,
) -> Stream:
    """
    Registers a delta stream with a target table version, replacing any
    previous stream registered for the same table version. Returns the
    committed stream.
    """
    raise NotImplementedError("commit_stream not implemented")


def delete_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    *args,
    **kwargs,
) -> None:
    """
    Deletes the delta stream currently registered with the given table version.
    Resolves to the latest active table version if no table version is given.
    Resolves to the deltacat stream format if no stream format is given.
    Raises an error if the stream does not exist.
    """
    raise NotImplementedError("delete_stream not implemented")


def delete_table(
    namespace: str,
    table_name: str,
    purge: bool = False,
    *args,
    **kwargs,
) -> None:
    """
    Drops the given table from the catalog. If purge is True, also removes
    all data files associated with the table. Raises an error if the given table
    does not exist.
    """
    raise NotImplementedError("delete_table not implemented")


def delete_namespace(
    namespace: str,
    purge: bool = False,
    *args,
    **kwargs,
) -> None:
    """
    Drops the given namespace from the catalog. If purge is True, also removes
    all data files associated with the namespace. Raises an error if the given
    namespace does not exist.
    """
    raise NotImplementedError("drop_namespace not implemented")


def get_stream_by_id(
    table_version_locator: TableVersionLocator,
    stream_id: str,
    *args,
    **kwargs,
) -> Optional[Partition]:
    """
    Gets the stream for the given table version locator and stream ID.
    Returns None if the stream does not exist. Raises an error if the given
    table version locator does not exist.
    """
    raise NotImplementedError("get_stream_by_id not implemented")


def get_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    *args,
    **kwargs,
) -> Optional[Stream]:
    """
    Gets the most recently committed stream for the given table version.
    Resolves to the latest active table version if no table version is given.
    Resolves to the deltacat stream format if no stream format is given.
    Returns None if the table version or stream format does not exist.
    """
    raise NotImplementedError("get_stream not implemented")


def stream_exists(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    *args,
    **kwargs,
) -> bool:
    """
    Returns True if the given Stream exists, False if not.
    Resolves to the latest active table version if no table version is given.
    Resolves to the DeltaCAT stream format if no stream format is given.
    Returns None if the table version or stream format does not exist.
    """
    raise NotImplementedError("stream_exists not implemented")


def stage_partition(
    stream: Stream,
    partition_values: Optional[PartitionValues] = None,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> Partition:
    """
    Stages a new partition for the given stream and partition values. Returns
    the staged partition. If this partition will replace another partition
    with the same partition values and scheme, then it will have its previous
    partition ID set to the ID of the partition being replaced. Partition values
    should not be specified for unpartitioned tables.

    The partition_values must represent the results of transforms in a partition
    spec specified in the stream.
    """
    raise NotImplementedError("stage_partition not implemented")


def commit_partition(
    partition: Partition,
    previous_partition: Optional[Partition] = None,
    expected_previous_partition_id: Optional[str] = UNKNOWN_PARTITION_ID,
    *args,
    **kwargs,
) -> Partition:
    """
    Commits the staged partition to its associated table version stream,
    replacing any previous partition registered for the same stream and
    partition values.

    If previous partition is given then it will be replaced with its deltas
    prepended to the new partition being committed. Otherwise the latest
    committed partition with the same keys and partition scheme ID will be
    retrieved.

    Returns the registered partition. If the partition's
    previous delta stream position is specified, then the commit will
    be rejected if it does not match the actual previous stream position of
    the partition being replaced. If the partition's previous partition ID is
    specified, then the commit will be rejected if it does not match the actual
    ID of the partition being replaced.
    """
    raise NotImplementedError("commit_partition not implemented")


def delete_partition(
    stream_locator: StreamLocator,
    partition_values: Optional[PartitionValues] = None,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    """
    Deletes the given partition from the specified stream. Partition
    values should not be specified for unpartitioned tables. Raises an error
    if the partition does not exist.
    """
    raise NotImplementedError("delete_partition not implemented")


def get_partition_by_id(
    stream_locator: StreamLocator,
    partition_id: str,
    *args,
    **kwargs,
) -> Optional[Partition]:
    """
    Gets the partition for the given stream locator and partition ID.
    Returns None if the partition does not exist. Raises an error if the
    given stream locator does not exist.
    """
    raise NotImplementedError("get_partition_by_id not implemented")


def get_partition(
    stream_locator: StreamLocator,
    partition_values: Optional[PartitionValues] = None,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Partition]:
    """
    Gets the most recently committed partition for the given stream locator and
    partition key values. Returns None if no partition has been committed for
    the given table version and/or partition key values. Partition values
    should not be specified for unpartitioned tables. Partition scheme ID
    resolves to the table version's current partition scheme by default.
    Raises an error if the given stream locator does not exist.
    """
    raise NotImplementedError("get_partition not implemented")


def stage_delta(
    data: Union[LocalTable, LocalDataset, DistributedDataset, Manifest],
    partition: Partition,
    delta_type: DeltaType = DeltaType.UPSERT,
    max_records_per_entry: Optional[int] = None,
    author: Optional[ManifestAuthor] = None,
    properties: Optional[DeltaProperties] = None,
    table_writer_kwargs: Optional[Dict[str, Any]] = None,
    content_type: ContentType = ContentType.PARQUET,
    entry_params: Optional[EntryParams] = None,
    entry_type: Optional[EntryType] = EntryType.DATA,
    schema: Optional[Schema] = None,
    sort_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> Delta:
    """
    Writes the given dataset to 1 or more files. Returns an unregistered
    delta whose manifest entries point to the uploaded files. Applies any
    schema consistency policies configured for the parent table version.
    """
    raise NotImplementedError("stage_delta not implemented")


def commit_delta(delta: Delta, *args, **kwargs) -> Delta:
    """
    Registers a new delta with its associated target table version and
    partition. Returns the registered delta. If the delta's previous stream
    position is specified, then the commit will be rejected if it does not match
    the target partition's actual previous stream position. If the delta's
    stream position is specified, it must be greater than the latest stream
    position in the target partition.
    """
    raise NotImplementedError("commit_delta not implemented")


def get_namespace(namespace: str, *args, **kwargs) -> Optional[Namespace]:
    """
    Gets table namespace metadata for the specified table namespace. Returns
    None if the given namespace does not exist.
    """
    raise NotImplementedError("get_namespace not implemented")


def namespace_exists(namespace: str, *args, **kwargs) -> bool:
    """
    Returns True if the given table namespace exists, False if not.
    """
    raise NotImplementedError("namespace_exists not implemented")


def get_table(namespace: str, table_name: str, *args, **kwargs) -> Optional[Table]:
    """
    Gets table metadata for the specified table. Returns None if the given
    table does not exist.
    """
    raise NotImplementedError("get_table not implemented")


def table_exists(namespace: str, table_name: str, *args, **kwargs) -> bool:
    """
    Returns True if the given table exists, False if not.
    """
    raise NotImplementedError("table_exists not implemented")


def get_table_version(
    namespace: str, table_name: str, table_version: str, *args, **kwargs
) -> Optional[TableVersion]:
    """
    Gets table version metadata for the specified table version. Returns None
    if the given table version does not exist.
    """
    raise NotImplementedError("get_table_version not implemented")


def get_latest_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:
    """
    Gets table version metadata for the latest version of the specified table.
    Returns None if no table version exists for the given table.
    """
    raise NotImplementedError("get_latest_table_version not implemented")


def get_latest_active_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:
    """
    Gets table version metadata for the latest active version of the specified
    table. Returns None if no active table version exists for the given table.
    """
    raise NotImplementedError("get_latest_active_table_version not implemented")


def get_table_version_column_names(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[List[str]]:
    """
    Gets a list of column names for the specified table version, or for the
    latest active table version if none is specified. The index of each
    column name returned represents its ordinal position in a delimited text
    file or other row-oriented content type files appended to the table.
    Returns None for schemaless tables. Raises an error if the table version
    does not exist.
    """
    raise NotImplementedError("get_table_version_column_names not implemented")


def get_table_version_schema(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Schema]:
    """
    Gets the schema for the specified table version, or for the latest active
    table version if none is specified. Returns None if the table version is
    schemaless. Raises an error if the table version does not exist.
    """
    raise NotImplementedError("get_table_version_schema not implemented")


def table_version_exists(
    namespace: str, table_name: str, table_version: str, *args, **kwargs
) -> bool:
    """
    Returns True if the given table version exists, False if not.
    """
    raise NotImplementedError("table_version_exists not implemented")


def can_categorize(e: BaseException, *args, **kwargs) -> bool:
    """
    True if the input error originated from the storage
    implementation layer and can be categorized under an
    existing DeltaCatError. The "categorize_errors" decorator
    uses this to determine if an unknown error from the storage
    implementation can be categorized prior to casting it to
    the equivalent DeltaCatError via `raise_categorized_error`
    """
    raise NotImplementedError


def raise_categorized_error(e: BaseException, *args, **kwargs):
    """
    Casts a categorizable error that originaed from the storage
    implementation layer to its equivalent DeltaCatError
    for uniform handling (e.g., determining whether an error
    is retryable or not) via the "categorize_errors" decorator.
    Raises an UnclassifiedDeltaCatError from the input exception
    if the error cannot be categorized.
    """
    raise NotImplementedError



import pytest
import pyarrow as pa
from deltacat.experimental.storage.rivulet import Schema, Field, Datatype


def test_field_initialization():
    field = Field(name="test_field", datatype=Datatype.string(), is_merge_key=True)
    assert field.name == "test_field"
    assert field.datatype == Datatype.string()
    assert field.is_merge_key


def test_schema_initialization():
    fields = [("id", Datatype.int64()), ("name", Datatype.string())]
    schema = Schema(fields, merge_keys=["id"])
    assert len(schema) == 2
    assert "id" in schema.keys()
    assert schema["id"].datatype == Datatype.int64()
    assert "name" in schema.keys()
    assert schema["name"].datatype == Datatype.string()


def test_merge_key_conflict_on_init():
    fields = [
        Field("id", Datatype.int64(), is_merge_key=False),  # Merge key off here
        ("name", Datatype.string()),
    ]
    with pytest.raises(TypeError):
        Schema(fields, merge_keys=["id"])  # Merge key on here


def test_simultaneous_duplicate_field():
    with pytest.raises(ValueError):
        Schema(
            [
                ("id", Datatype.int32()),
                ("name", Datatype.string()),
                ("age", Datatype.int32()),
                ("age", Datatype.string()),
            ],
            merge_keys=["id"],
        )


def test_add_field():
    schema = Schema()
    field = Field("new_field", Datatype.float(), True)
    schema.add_field(field)
    assert len(schema) == 1
    assert "new_field" in schema.keys()
    assert schema["new_field"].datatype == Datatype.float()

    field2 = Field("another_field", Datatype.string(), True)
    schema.add_field(field2)
    assert len(schema) == 2
    assert "another_field" in schema.keys()
    assert schema["another_field"].datatype == Datatype.string()

    with pytest.raises(ValueError):
        schema.add_field(field2)


def test_setitem_field():
    schema = Schema()
    field = Field("test_field", Datatype.int64(), is_merge_key=True)
    schema["test_field"] = field
    assert schema["test_field"] == field


def test_setitem_datatype():
    schema = Schema()
    schema["id"] = (Datatype.int64(), True)
    schema["test_field"] = Datatype.int64()
    assert schema["test_field"].name == "test_field"
    assert schema["test_field"].datatype == Datatype.int64()
    assert not schema["test_field"].is_merge_key


def test_setitem_tuple_with_merge_key():
    schema = Schema()
    schema["test_field"] = (Datatype.int64(), True)
    assert schema["test_field"].name == "test_field"
    assert schema["test_field"].datatype == Datatype.int64()
    assert schema["test_field"].is_merge_key


def test_setitem_invalid_type():
    schema = Schema()
    with pytest.raises(TypeError):
        schema["test_field"] = "invalid"


def test_non_empty_merge_key():
    with pytest.raises(TypeError):
        _ = Schema([], merge_keys=["id"])


def test_merge_schemas():
    schema1 = Schema([("id", Datatype.int64())], merge_keys=["id"])
    schema2 = Schema(
        [("other_id", Datatype.string()), ("name", Datatype.string())],
        merge_keys="other_id",
    )
    schema1.merge(schema2)
    assert len(schema1) == 3
    assert "id" in schema1.keys()
    assert "name" in schema1.keys()
    assert "other_id" in schema1.keys()


def test_merge_schemas_same_merge_key():
    schema1 = Schema(
        [("id", Datatype.int64()), ("name", Datatype.string())], merge_keys=["id"]
    )
    schema2 = Schema(
        [("id", Datatype.int64()), ("other_name", Datatype.string())],
        merge_keys="id",
    )
    schema1.merge(schema2)
    assert len(schema1) == 3
    assert "id" in schema1.keys()
    assert "name" in schema1.keys()
    assert "other_name" in schema1.keys()


def test_merge_schema_conflict():
    schema1 = Schema([("id", Datatype.int64())], merge_keys=["id"])
    schema1_dup = Schema([("id", Datatype.int64())], merge_keys=["id"])
    schema2 = Schema([("id", Datatype.string())], merge_keys=["id"])

    with pytest.raises(ValueError):
        schema1.merge(schema2)

    schema1.merge(
        schema1_dup
    )  # Merging the same field is allowed (unlike using add_field)
    assert schema1["id"].datatype == Datatype.int64()
    assert len(schema1) == 1


def test_to_pyarrow_schema():
    fields = [("id", Datatype.int64()), ("name", Datatype.string())]
    schema = Schema(fields, merge_keys=["id"])
    pa_schema = schema.to_pyarrow()
    assert isinstance(pa_schema, pa.Schema)
    assert len(pa_schema) == 2
    assert pa_schema.field("id").type == pa.int64()
    assert pa_schema.field("name").type == pa.string()


def test_from_pyarrow_schema():
    pa_schema = pa.schema([("id", pa.int64()), ("name", pa.string())])
    schema = Schema.from_pyarrow(pa_schema, merge_keys=["id"])
    assert len(schema) == 2
    assert schema["id"].is_merge_key


def test_from_pyarrow_schema_invalid_merge_keys():
    pa_schema = pa.schema([("id", pa.int64()), ("name", pa.string())])
    with pytest.raises(ValueError):
        Schema.from_pyarrow(pa_schema, merge_keys=["bad_key"])


def test_get_field():
    schema = Schema([("id", Datatype.int64())], merge_keys=["id"])
    field = schema["id"]
    assert field.name == "id"
    assert field.datatype == Datatype.int64()


def test_set_field():
    schema = Schema([("id", Datatype.int64())], merge_keys=["id"])
    schema["name"] = Field("name", Datatype.string())
    assert len(schema) == 2
    assert "name" in schema.keys()
    assert schema["name"].datatype == Datatype.string()


def test_delete_field():
    schema = Schema(
        [("name", Datatype.string()), ("zip", Datatype.int32())], merge_keys=["name"]
    )
    del schema["zip"]
    assert "zip" not in schema.keys()
    assert "name" in schema.keys()


def test_delete_merge_key_field():
    schema = Schema([("id", Datatype.int64())], merge_keys=["id"])
    with pytest.raises(ValueError):
        del schema["id"]


def test_schema_iter():
    fields = [
        Field("id", Datatype.int32(), is_merge_key=True),
        Field("name", Datatype.string()),
    ]
    schema = Schema(fields)
    iter_result = list(iter(schema))
    assert len(iter_result) == 2
    assert all(isinstance(item, str) for item in iter_result)


def test_merge_all():
    schema1 = Schema(
        [
            Field("id", Datatype.int64(), is_merge_key=True),
            Field("name", Datatype.string()),
        ]
    )
    schema2 = Schema(
        [
            Field("age", Datatype.int32()),
            Field("email", Datatype.string(), is_merge_key=True),
        ]
    )
    merged_schema = Schema.merge_all([schema1, schema2])
    assert len(merged_schema) == 4


def test_schema_values():
    fields = [
        Field("id", Datatype.int64(), is_merge_key=True),
        Field("name", Datatype.string()),
    ]
    schema = Schema(fields)
    values = list(schema.values())
    assert len(values) == 2
    assert all(isinstance(v, Field) for v in values)


def test_schema_items():
    fields = [
        Field("id", Datatype.int64(), is_merge_key=True),
        Field("name", Datatype.string()),
    ]
    schema = Schema(fields)
    items = list(schema.items())
    assert len(items) == 2
    assert all(isinstance(k, str) and isinstance(v, Field) for k, v in items)



import unittest
import json
import ray
from unittest import mock
from logging import LogRecord
from deltacat.logs import JsonFormatter


class TestJsonFormatter(unittest.TestCase):
    def test_usesTime_sanity(self):

        formatter = JsonFormatter()

        self.assertFalse(formatter.usesTime())

    def test_usesTime_success_case(self):

        formatter = JsonFormatter(fmt_dict={"asctime": "asctime"})

        self.assertTrue(formatter.usesTime())

    def test_formatMessage_sanity(self):

        formatter = JsonFormatter({"message": "msg"})

        record = LogRecord(
            level="INFO",
            name="test",
            pathname="test",
            lineno=0,
            message="test_message",
            msg="test_message",
            args=None,
            exc_info=None,
        )

        result = formatter.formatMessage(record)

        self.assertEqual({"message": "test_message"}, result)

    def test_format_sanity(self):
        ray.shutdown()
        formatter = JsonFormatter({"message": "msg"})

        record = LogRecord(
            level="INFO",
            name="test",
            pathname="test",
            lineno=0,
            message="test_message",
            msg="test_message",
            args=None,
            exc_info=None,
        )

        result = formatter.format(record)

        self.assertEqual({"message": "test_message"}, json.loads(result))
        self.assertFalse(ray.is_initialized())
        self.assertNotIn("ray_runtime_context", json.loads(result))

    def test_format_when_ray_initialized(self):
        ray.init(local_mode=True, ignore_reinit_error=True)

        formatter = JsonFormatter({"message": "msg"})

        record = LogRecord(
            level="INFO",
            name="test",
            pathname="test",
            lineno=0,
            message="test_message",
            msg="test_message",
            args=None,
            exc_info=None,
        )

        result = formatter.format(record)
        result = json.loads(result)

        self.assertEqual("test_message", result["message"])
        self.assertTrue(ray.is_initialized())
        self.assertIn("ray_runtime_context", result)
        self.assertIn("job_id", result["ray_runtime_context"])
        self.assertIn("node_id", result["ray_runtime_context"])
        self.assertIn("worker_id", result["ray_runtime_context"])
        self.assertNotIn(
            "task_id",
            result["ray_runtime_context"],
            "We expect task ID not be present outside a remote task",
        )
        ray.shutdown()

    def test_format_when_ray_initialized_in_task(self):
        # worker mode is only true when local_mode is False
        ray.init(local_mode=False, ignore_reinit_error=True)

        @ray.remote
        def ray_remote_task():
            formatter = JsonFormatter({"message": "msg"})

            record = LogRecord(
                level="INFO",
                name="test",
                pathname="test",
                lineno=0,
                message="test_message",
                msg="test_message",
                args=None,
                exc_info=None,
            )

            result = formatter.format(record)
            result = json.loads(result)
            return result

        result = ray.get(ray_remote_task.remote())
        self.assertEqual("test_message", result["message"])
        self.assertTrue(ray.is_initialized())
        self.assertIn("ray_runtime_context", result)
        self.assertIn("job_id", result["ray_runtime_context"])
        self.assertIn("node_id", result["ray_runtime_context"])
        self.assertIn("worker_id", result["ray_runtime_context"])
        self.assertIn(
            "task_id",
            result["ray_runtime_context"],
            "We expect task ID to be present inside a remote task",
        )
        ray.shutdown()

    def test_format_when_ray_initialized_with_context_kwargs(self):
        ray.init(local_mode=True, ignore_reinit_error=True)

        formatter = JsonFormatter(
            {"message": "msg"}, context_kwargs={"custom_key": "custom_val"}
        )

        record = LogRecord(
            level="INFO",
            name="test",
            pathname="test",
            lineno=0,
            message="test_message",
            msg="test_message",
            args=None,
            exc_info=None,
        )

        result = formatter.format(record)
        result = json.loads(result)

        self.assertEqual("test_message", result["message"])
        self.assertTrue(ray.is_initialized())
        self.assertIn("ray_runtime_context", result)
        self.assertIn("job_id", result["ray_runtime_context"])
        self.assertIn("node_id", result["ray_runtime_context"])
        self.assertIn("worker_id", result["ray_runtime_context"])
        self.assertNotIn(
            "task_id",
            result["ray_runtime_context"],
            "We expect task ID not be present outside a remote task",
        )
        self.assertEqual("custom_val", result["additional_context"]["custom_key"])
        ray.shutdown()

    def test_format_with_context_kwargs(self):
        ray.shutdown()
        formatter = JsonFormatter(
            {"message": "msg"}, context_kwargs={"custom_key": "custom_val"}
        )

        record = LogRecord(
            level="INFO",
            name="test",
            pathname="test",
            lineno=0,
            message="test_message",
            msg="test_message",
            args=None,
            exc_info=None,
        )

        result = formatter.format(record)

        self.assertEqual(
            {
                "message": "test_message",
                "additional_context": {"custom_key": "custom_val"},
            },
            json.loads(result),
        )
        self.assertFalse(ray.is_initialized())
        self.assertNotIn("ray_runtime_context", json.loads(result))

    @mock.patch("deltacat.logs.DELTACAT_LOGGER_CONTEXT", '{"DATABASE_URL": "mytemp"}')
    def test_format_with_env_context_kwargs(self):
        ray.shutdown()
        formatter = JsonFormatter(
            {"message": "msg"}, context_kwargs={"custom_key": "custom_val"}
        )

        record = LogRecord(
            level="INFO",
            name="test",
            pathname="test",
            lineno=0,
            message="test_message",
            msg="test_message",
            args=None,
            exc_info=None,
        )

        result = formatter.format(record)

        self.assertEqual(
            {
                "message": "test_message",
                "additional_context": {
                    "custom_key": "custom_val",
                    "DATABASE_URL": "mytemp",
                },
            },
            json.loads(result),
        )
        self.assertFalse(ray.is_initialized())
        self.assertNotIn("ray_runtime_context", json.loads(result))



import logging
from typing import Any

import pyarrow
from deltacat.storage.model.scan.push_down import PartitionFilter

import deltacat.logs as logs
from deltacat.storage.model.expression import Reference, Literal
from deltacat.storage.model.expression.visitor import ExpressionVisitor
from pyiceberg.expressions import (
    And,
    Or,
    Not,
    EqualTo,
    NotEqualTo,
    GreaterThan,
    GreaterThanOrEqual,
    LessThan,
    LessThanOrEqual,
    IsNull,
    In,
)

# Initialize DeltaCAT logger
logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class IcebergExpressionVisitor(ExpressionVisitor[None, Any]):
    """
    Visitor that translates DeltaCAT expressions to PyIceberg expressions.
    """

    def visit(self, expr, context=None):
        # Handle PartitionFilter by extracting and visiting the inner expression
        if isinstance(expr, PartitionFilter):
            return self.visit(expr.expr, context)
        # Handle all other expressions using the parent's visit method
        return super().visit(expr, context)

    def visit_reference(self, expr: Reference, context=None) -> str:
        return expr.field

    def visit_literal(self, expr: Literal, context=None) -> Any:
        # Convert PyArrow scalar to Python native type
        return (
            expr.value.as_py() if isinstance(expr.value, pyarrow.Scalar) else expr.value
        )

    def visit_and(self, expr, context=None):
        left = self.visit(expr.left, context)
        right = self.visit(expr.right, context)
        return And(left, right)

    def visit_or(self, expr, context=None):
        left = self.visit(expr.left, context)
        right = self.visit(expr.right, context)
        return Or(left, right)

    def visit_not(self, expr, context=None):
        operand = self.visit(expr.operand, context)
        return Not(operand)

    def visit_equal(self, expr, context=None):
        left = self.visit(expr.left, context)
        right = self.visit(expr.right, context)
        return EqualTo(left, right)

    def visit_not_equal(self, expr, context=None):
        left = self.visit(expr.left, context)
        right = self.visit(expr.right, context)
        return NotEqualTo(left, right)

    def visit_greater_than(self, expr, context=None):
        left = self.visit(expr.left, context)
        right = self.visit(expr.right, context)
        return GreaterThan(left, right)

    def visit_greater_than_equal(self, expr, context=None):
        left = self.visit(expr.left, context)
        right = self.visit(expr.right, context)
        return GreaterThanOrEqual(left, right)

    def visit_less_than(self, expr, context=None):
        left = self.visit(expr.left, context)
        right = self.visit(expr.right, context)
        return LessThan(left, right)

    def visit_less_than_equal(self, expr, context=None):
        left = self.visit(expr.left, context)
        right = self.visit(expr.right, context)
        return LessThanOrEqual(left, right)

    def visit_is_null(self, expr, context=None):
        operand = self.visit(expr.operand, context)
        return IsNull(operand)

    def visit_in(self, expr, context=None):
        value = self.visit(expr.value, context)
        values = [self.visit(v, context) for v in expr.values]
        return In(value, values)

    def visit_between(self, expr, context=None):
        value = self.visit(expr.value, context)
        lower = self.visit(expr.lower, context)
        upper = self.visit(expr.upper, context)
        return And(GreaterThanOrEqual(value, lower), LessThanOrEqual(value, upper))

    # PyIceberg does not have a direct equivalent of LIKE
    def visit_like(self, expr, context=None):
        value = self.visit(expr.value, context)
        pattern = self.visit(expr.pattern, context)
        logger.warning(
            f"LIKE operation is not supported in PyIceberg. Ignoring LIKE filter: {value} LIKE '{pattern}'. "
            "This may result in more data being returned than expected."
        )
        # Return None or a default expression that won't filter anything
        return (
            None  # or return NotEqualTo(value, None)  # matches everything except NULL
        )



import logging
from typing import Callable, Dict, List, Optional, Union

from fsspec import AbstractFileSystem

import pyarrow as pa
from pyarrow import csv as pacsv
import pyarrow.fs as pafs

from ray.data import Dataset
from ray.data.datasource import FilenameProvider

from deltacat import logs
from deltacat.types.media import ContentEncoding, ContentType

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def write_parquet(
    dataset: Dataset,
    base_path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]],
    block_path_provider: Union[Callable, FilenameProvider],
    **kwargs,
) -> None:

    dataset.write_parquet(
        base_path,
        filesystem=filesystem,
        try_create_dir=False,
        filename_provider=block_path_provider,
        **kwargs,
    )


def write_csv(
    dataset: Dataset,
    base_path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]],
    block_path_provider: Union[Callable, FilenameProvider],
    **kwargs,
) -> None:
    """
    Write a Ray Dataset to a CSV file (or other delimited text format).
    """
    # Extract CSV-specific options from kwargs
    delimiter = kwargs.pop("delimiter", ",")
    quoting_style = kwargs.pop("quoting_style", None)
    include_header = kwargs.pop("include_header", False)

    # Create a function that will generate WriteOptions inside the worker process
    def arrow_csv_args_fn():
        write_options = pacsv.WriteOptions(
            delimiter=delimiter,
            include_header=include_header,
            quoting_style=quoting_style,
        )
        return {"write_options": write_options}

    # Check if the block_path_provider will generate .gz files to avoid double compression
    pa_open_stream_args = {}
    if not (
        hasattr(block_path_provider, "content_encoding")
        and block_path_provider.content_encoding == ContentEncoding.GZIP
    ):
        # Block path provider will not generate .gz files, so we need to apply explicit compression
        pa_open_stream_args["compression"] = ContentEncoding.GZIP.value

    dataset.write_csv(
        base_path,
        arrow_open_stream_args=pa_open_stream_args,
        filesystem=filesystem,
        try_create_dir=False,
        filename_provider=block_path_provider,
        arrow_csv_args_fn=arrow_csv_args_fn,
        **kwargs,
    )


def write_json(
    dataset: Dataset,
    base_path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]],
    block_path_provider: Union[Callable, FilenameProvider],
    **kwargs,
) -> None:
    """
    Write a Ray Dataset to a JSON file using Ray's native JSON writer.
    """
    # Check if the block_path_provider will generate .gz files to avoid double compression
    pa_open_stream_args = {}
    if not (
        hasattr(block_path_provider, "content_encoding")
        and block_path_provider.content_encoding == ContentEncoding.GZIP
    ):
        # Block path provider will not generate .gz files, so we need to apply explicit compression
        pa_open_stream_args["compression"] = ContentEncoding.GZIP.value

    dataset.write_json(
        base_path,
        arrow_open_stream_args=pa_open_stream_args,
        filesystem=filesystem,
        try_create_dir=False,
        filename_provider=block_path_provider,
        **kwargs,
    )


CONTENT_TYPE_TO_DATASET_WRITE_FUNC: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV.value: write_csv,
    ContentType.TSV.value: write_csv,
    ContentType.CSV.value: write_csv,
    ContentType.PSV.value: write_csv,
    ContentType.PARQUET.value: write_parquet,
    ContentType.JSON.value: write_json,
}


def content_type_to_writer_kwargs(content_type: str) -> Dict[str, any]:
    """
    Returns writer kwargs for the given content type when writing with Ray Dataset.
    """
    if content_type == ContentType.UNESCAPED_TSV.value:
        return {
            "delimiter": "\t",
            "include_header": False,
            "quoting_style": "none",
        }
    if content_type == ContentType.TSV.value:
        return {
            "delimiter": "\t",
            "include_header": False,
        }
    if content_type == ContentType.CSV.value:
        return {
            "delimiter": ",",
            "include_header": False,
        }
    if content_type == ContentType.PSV.value:
        return {
            "delimiter": "|",
            "include_header": False,
        }
    if content_type in {ContentType.PARQUET.value, ContentType.JSON.value}:
        return {}
    raise ValueError(f"Unsupported content type: {content_type}")


def slice_dataset(dataset: Dataset, max_len: Optional[int]) -> List[Dataset]:
    """
    Returns equally-sized dataset slices of up to `max_len` records each.
    """
    # TODO (pdames): Since the purpose of this function is to ideally produce
    #  a dataset that won't write less than `max_len` records per file when the
    #  total dataset size is greater than `max_len`, we should also ensure that
    #  each dataset returned only contains 1 block.
    if max_len is None:
        return [dataset]
    dataset_len = dataset.count()
    num_splits = int(dataset_len / max_len)
    if dataset_len % max_len == 0:
        num_splits -= 1
    if num_splits <= 0:
        return [dataset]
    split_indices = [max_len + max_len * i for i in range(num_splits)]
    return dataset.split_at_indices(split_indices)


def dataset_size(dataset: Dataset) -> int:
    return dataset.size_bytes()


def dataset_to_file(
    table: Dataset,
    base_path: str,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]],
    block_path_provider: Union[Callable, FilenameProvider],
    content_type: str = ContentType.PARQUET.value,
    schema: Optional[pa.Schema] = None,
    **kwargs,
) -> None:
    """
    Writes the given Distributed Dataset to one or more files.
    """
    writer = CONTENT_TYPE_TO_DATASET_WRITE_FUNC.get(content_type)
    if not writer:
        raise NotImplementedError(
            f"Distributed Dataset writer for content type '{content_type}' not"
            f" implemented. Known content types: "
            f"{CONTENT_TYPE_TO_DATASET_WRITE_FUNC.keys}"
        )
    writer_kwargs = content_type_to_writer_kwargs(content_type)
    writer_kwargs.update(kwargs)
    writer(
        table,
        base_path,
        filesystem=filesystem,
        block_path_provider=block_path_provider,
        **writer_kwargs,
    )



from contextlib import contextmanager
import io
from abc import ABC, abstractmethod
from typing import Protocol

from pyarrow.fs import FileSystem, FileType


class InputStream(Protocol):
    """A protocol with a subset of IOBase for file-like input objects"""

    @abstractmethod
    def read(self, size: int = -1) -> bytes:
        ...

    @abstractmethod
    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        ...

    @abstractmethod
    def tell(self) -> int:
        ...

    @abstractmethod
    def close(self) -> None:
        ...

    def __enter__(self) -> "InputStream":
        ...

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...


class InputFile(ABC):
    """Abstraction for interacting with input files"""

    def __init__(self, location: str):
        self._location = location

    @property
    def location(self) -> str:
        return self._location

    @abstractmethod
    def exists(self) -> bool:
        """Return whether the location exists.

        Raises:
            PermissionError: If this has insufficient permissions to access the file at location.
        """

    @abstractmethod
    def open(self) -> InputStream:
        """Return a file-like object for input

        Raises:
            FileNotFoundError: If the file does not exist at self.location.
            PermissionError: If this has insufficient permissions to access the file at location.
        """


class FSInputFile(InputFile):
    def __init__(self, location: str, fs: FileSystem):
        self._location = location
        self.fs = fs

    def exists(self) -> bool:
        file_info = self.fs.get_file_info(self._location)
        return file_info.type != FileType.NotFound

    @contextmanager
    def open(self, seekable: bool = False):
        with self.fs.open_input_file(self._location) as input_stream:
            yield input_stream



from deltacat.io.reader.deltacat_read_api import read_deltacat
from deltacat.io.datasource.deltacat_datasource import DeltacatReadType
from deltacat.io.datasource.deltacat_datasource import (
    METAFILE_DATA_COLUMN_NAME,
    METAFILE_TYPE_COLUMN_NAME,
)

__all__ = [
    "read_deltacat",
    "DeltacatReadType",
    "METAFILE_DATA_COLUMN_NAME",
    "METAFILE_TYPE_COLUMN_NAME",
]



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import logging
import time
from abc import ABC, abstractmethod
from collections import defaultdict

from deltacat.utils.common import ReadKwargsProvider
from ray.types import ObjectRef

from deltacat.compute.compactor.model.delta_file_envelope import DeltaFileEnvelopeGroups
from deltacat.compute.compactor_v2.utils.delta import read_delta_file_envelopes

from deltacat.compute.compactor_v2.utils.primary_key_index import (
    hash_group_index_to_hash_bucket_indices,
)

from deltacat.storage import metastore

from deltacat.io.object_store import IObjectStore

from deltacat import logs

from deltacat.compute.compactor import DeltaFileEnvelope, DeltaAnnotated

from typing import List, Optional

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class MergeFileGroup(dict):
    @staticmethod
    def of(hb_index: int, dfe_groups: Optional[List[List[DeltaFileEnvelope]]] = None):
        """
        Creates a container with delta file envelope groupings and other
        additional properties used primarily for the merging step.

        Args:
            hb_index: This signifies the hash bucket index corresponding to the envelope delta file groups.
            dfe_groups: A list of delta file envelope groups.
                If not present, the provided hash bucket index is a copy by reference candidate during the merge step.

        Returns:
            A dict

        """
        d = MergeFileGroup()
        d["hb_index"] = hb_index
        d["dfe_groups"] = dfe_groups
        return d

    @property
    def dfe_groups(self) -> Optional[List[List[DeltaFileEnvelope]]]:
        return self["dfe_groups"]

    @property
    def hb_index(self) -> int:
        return self["hb_index"]


class MergeFileGroupsProvider(ABC):
    @abstractmethod
    def create(self) -> List[MergeFileGroup]:
        """
        Creates a list of merge file groups.

        Returns: a list of merge file groups.

        """
        raise NotImplementedError("Method not implemented")

    @property
    @abstractmethod
    def hash_group_index(self):
        raise NotImplementedError("Method not implemented")


class LocalMergeFileGroupsProvider(MergeFileGroupsProvider):
    """
    A factory class for producing merge file groups given local delta file envelopes.
    """

    LOCAL_HASH_BUCKET_INDEX = 0
    LOCAL_HASH_GROUP_INDEX = 0

    def __init__(
        self,
        uniform_deltas: List[DeltaAnnotated],
        all_column_names: List[str],
        read_kwargs_provider: Optional[ReadKwargsProvider],
        deltacat_storage=metastore,
        deltacat_storage_kwargs: Optional[dict] = None,
    ):
        self._deltas = uniform_deltas
        self._all_column_names = all_column_names
        self._read_kwargs_provider = read_kwargs_provider
        self._deltacat_storage = deltacat_storage
        self._deltacat_storage_kwargs = deltacat_storage_kwargs
        self._loaded_deltas = False

    def _read_deltas_locally(self):
        local_dfe_list = []
        input_records_count = 0
        uniform_deltas = self._deltas
        logger.info(f"Getting {len(uniform_deltas)} DFE Tasks.")
        dfe_start = time.monotonic()
        for annotated_delta in uniform_deltas:
            (
                delta_file_envelopes,
                total_record_count,
                total_size_bytes,
            ) = read_delta_file_envelopes(
                annotated_delta,
                self._all_column_names,
                self._read_kwargs_provider,
                self._deltacat_storage,
                self._deltacat_storage_kwargs,
            )
            if delta_file_envelopes:
                local_dfe_list.extend(delta_file_envelopes)
                input_records_count += total_record_count
        dfe_end = time.monotonic()
        logger.info(
            f"Retrieved {len(local_dfe_list)} DFE Tasks in {dfe_end - dfe_start}s."
        )

        self._dfe_groups = [local_dfe_list] if len(local_dfe_list) > 0 else None
        self._loaded_deltas = True

    def create(self) -> List[MergeFileGroup]:
        if not self._loaded_deltas:
            self._read_deltas_locally()

        # Since hash bucketing is skipped for local merges, we use a fixed index here.
        return [
            MergeFileGroup.of(
                hb_index=LocalMergeFileGroupsProvider.LOCAL_HASH_BUCKET_INDEX,
                dfe_groups=self._dfe_groups,
            )
        ]

    @property
    def hash_group_index(self):
        return LocalMergeFileGroupsProvider.LOCAL_HASH_GROUP_INDEX


class RemoteMergeFileGroupsProvider(MergeFileGroupsProvider):
    """
    A factory class for producing merge file groups given delta file envelope object refs
        and hash bucketing parameters. Delta file envelopes are pulled from the object store
        remotely and loaded with in-memory pyarrow tables.
    """

    def __init__(
        self,
        hash_group_index: int,
        dfe_groups_refs: List[ObjectRef[DeltaFileEnvelopeGroups]],
        hash_bucket_count: int,
        num_hash_groups: int,
        object_store: IObjectStore,
    ):
        self.hash_bucket_count = hash_bucket_count
        self.num_hash_groups = num_hash_groups
        self.object_store = object_store
        self._hash_group_index = hash_group_index
        self._dfe_groups_refs = dfe_groups_refs
        self._dfe_groups = []
        self._loaded_from_object_store = False

    def _load_deltas_from_object_store(self):
        delta_file_envelope_groups_list = self.object_store.get_many(
            self._dfe_groups_refs
        )
        hb_index_to_delta_file_envelopes_list = defaultdict(list)
        for delta_file_envelope_groups in delta_file_envelope_groups_list:
            assert self.hash_bucket_count == len(delta_file_envelope_groups), (
                f"The hash bucket count must match the dfe size as {self.hash_bucket_count}"
                f" != {len(delta_file_envelope_groups)}"
            )

            for hb_idx, dfes in enumerate(delta_file_envelope_groups):
                if dfes:
                    hb_index_to_delta_file_envelopes_list[hb_idx].append(dfes)
        valid_hb_indices_iterable = hash_group_index_to_hash_bucket_indices(
            self.hash_group_index, self.hash_bucket_count, self.num_hash_groups
        )

        total_dfes_found = 0
        dfe_list_groups = []
        for hb_idx in valid_hb_indices_iterable:
            dfe_list = hb_index_to_delta_file_envelopes_list.get(hb_idx)
            if dfe_list:
                total_dfes_found += 1
                dfe_list_groups.append(
                    MergeFileGroup.of(hb_index=hb_idx, dfe_groups=dfe_list)
                )
            else:
                dfe_list_groups.append(MergeFileGroup.of(hb_index=hb_idx))

        assert total_dfes_found == len(hb_index_to_delta_file_envelopes_list), (
            "The total dfe list does not match the input dfes from hash bucket as "
            f"{total_dfes_found} != {len(hb_index_to_delta_file_envelopes_list)}"
        )
        self._dfe_groups = dfe_list_groups
        self._loaded_from_object_store = True

    def create(self) -> List[MergeFileGroup]:
        if not self._loaded_from_object_store:
            self._load_deltas_from_object_store()

        return self._dfe_groups

    @property
    def hash_group_index(self):
        return self._hash_group_index



import logging

from collections import OrderedDict
from typing import Dict, Any, Optional, List, Iterable

from ray.data import Datasink
from ray.data._internal.execution.interfaces import TaskContext
from ray.data.block import Block, BlockAccessor
from ray.data.datasource import WriteResult

from ray.data.datasource.filename_provider import (
    FilenameProvider,
)

from deltacat import logs

from deltacat.constants import METAFILE_FORMAT_MSGPACK
from deltacat.storage import Metafile
from deltacat.io.datasource.deltacat_datasource import (
    METAFILE_DATA_COLUMN_NAME,
    METAFILE_TYPE_COLUMN_NAME,
)
from deltacat.utils.url import DeltaCatUrl, DeltaCatUrlWriter

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class CapturingBlockWritePathProvider(FilenameProvider):
    """Delegating block write path provider that saves an ordered dictionary of
    input keyword arguments for every block write path returned."""

    def __init__(
        self,
        block_write_path_provider: FilenameProvider,
        base_path: Optional[str] = None,
    ):
        self.base_path = base_path
        self.block_write_path_provider = block_write_path_provider
        self.write_path_kwargs: Dict[str, Dict[str, Any]] = OrderedDict()

    def get_filename_for_block(
        self,
        block: Any,
        task_index: int,
        block_index: int,
    ) -> str:
        if self.base_path is None:
            raise ValueError(
                "Base path must be provided to CapturingBlockWritePathProvider",
            )
        return self._get_write_path_for_block(
            base_path=self.base_path,
            block=block,
            block_index=block_index,
        )

    def _get_write_path_for_block(
        self,
        base_path: str,
        *args,
        **kwargs,
    ) -> str:
        filename = self.block_write_path_provider.get_filename_for_block(
            *args,
            **kwargs,
        )
        write_path = f"{base_path}/{filename}"
        kwargs["base_path"] = base_path
        self.write_path_kwargs[write_path] = kwargs
        return write_path


class DeltaCatWriteResult:
    def __init__(self):
        self.metadata = None
        self.path = None
        self.dataset_uuid = None
        self.block_write_path_provider = None
        self.content_type = None
        self.content_encoding = None
        self.filesystem = None


class DeltaCatDatasink(Datasink[List[Metafile]]):
    def __init__(
        self,
        url: DeltaCatUrl,
        *,
        metadata_only: bool = False,
        copy_on_write: Optional[bool] = False,
    ):
        self._url = url
        self._metadata_only = metadata_only
        self._copy_on_write = copy_on_write

    def on_write_start(self) -> None:
        pass

    def write(
        self,
        blocks: Iterable[Block],
        ctx: TaskContext,
    ) -> List[Metafile]:
        for block in blocks:
            pa_table = BlockAccessor.for_block(block).to_arrow()
            if (
                METAFILE_DATA_COLUMN_NAME in pa_table.column_names
                and METAFILE_TYPE_COLUMN_NAME in pa_table.column_names
            ):
                for pa_scalar in pa_table[METAFILE_DATA_COLUMN_NAME]:
                    metafile_msgpack_bytes = pa_scalar.as_py()
                    metafile = Metafile.deserialize(
                        serialized=metafile_msgpack_bytes,
                        meta_format=METAFILE_FORMAT_MSGPACK,
                    )
                    # TODO(pdames): Add `metafile` to writer as a kwarg instead
                    #  of constructing a new URL with the metafile as input.
                    writer_url = DeltaCatUrlWriter(self._url, metafile=metafile)
                    # TODO(pdames): Run writes in order from catalog -> delta
                    #  by truncating the URL down to just dc://{catalog-name}
                    #  and rebuilding all path elements from there.
                    writer_url.write(metafile)
            else:
                raise NotImplementedError(
                    f"Expected {METAFILE_DATA_COLUMN_NAME} and "
                    f"{METAFILE_TYPE_COLUMN_NAME} columns in the input block, "
                    f"but found {pa_table.column_names}."
                )

    def on_write_complete(
        self,
        write_result: WriteResult[List[Metafile]],
    ):
        pass


"""
    def write(
        self,
        blocks: Iterable[Block],
        ctx: TaskContext,
    ) -> List[ObjectRef[DeltacatWriteResult]]:
        paths, filesystem = resolve_paths_and_filesystem(
            self.path,
            self.filesystem,
        )
        assert len(paths) == 1, f"Expected 1 write path, found {len(paths)}."
        path = paths[0]
        write_results = super().write(blocks)
        # append a summary of this write operation in the last write result
        metadata = [BlockAccessor.for_block(_).get_metadata() for _ in blocks]
        rwr = DeltacatWriteResult()
        rwr.metadata = metadata
        rwr.path = path
        rwr.dataset_uuid = self.dataset_uuid
        rwr.block_write_path_provider = self.filename_provider
        rwr.content_type = ContentType.PARQUET.value
        rwr.content_encoding = ContentEncoding.IDENTITY.value
        rwr.filesystem = filesystem
        rwr_obj_ref = ray.put(rwr)
        write_results.append(rwr_obj_ref)
        return write_results

    def on_write_complete(self, write_results: List[Any], **kwargs) -> None:
        # TODO (pdames): time latency of this operation - overall s3 write times
        #  are 2-3x pure read_parquet_fast() times
        # restore the write operation summary from the last write result
        result: DeltacatWriteResult = write_results[len(write_results) - 1]
        write_path_args = result.block_write_path_provider.write_path_kwargs
        blocks_written = len(write_path_args)
        expected_blocks_written = len(result.metadata)
        # TODO(pdames): Corner cases where mismatch is expected? Emply blocks?
        #  Blocks filtered/split/merged to more/less write paths?
        assert blocks_written == expected_blocks_written, (
            f"Dataset write result validation failed. Found "
            f"{blocks_written}/{expected_blocks_written} Dataset blocks "
            f"written. Refusing to commit DeltaCAT Manifest."
        )
        manifest_entries = ManifestEntryList()
        for block_idx, path in enumerate(write_path_args.keys()):
            file_info = result.filesystem.get_file_info(path)
            if file_info.type == pyarrow.fs.FileType.File:
                content_length = file_info.size
            else:
                raise FileNotFoundError(ENOENT, strerror(ENOENT), path)
            num_rows = result.metadata[block_idx].num_rows
            source_content_length = result.metadata[block_idx].size_bytes
            manifest_entry_meta = ManifestMeta.of(
                int(num_rows) if num_rows is not None else None,
                int(content_length) if content_length is not None else None,
                result.content_type,
                result.content_encoding,
                int(source_content_length) if source_content_length else None,
            )
            parsed_url = parse_s3_url(path)
            manifest_entry = ManifestEntry.of(
                parsed_url.url,
                manifest_entry_meta,
            )
            manifest_entries.append(manifest_entry)
        manifest = Manifest.of(manifest_entries)
        manifest_path = f"{result.path}/manifest"
        logger.debug(f"Write succeeded for Dataset ID: {result.dataset_uuid}")
        with result.filesystem.open_output_stream(manifest_path) as f:
            f.write(json.dumps(manifest).encode("utf-8"))
        logger.debug(f"Manifest committed to: {manifest_path}")
"""



from abc import ABC, abstractmethod
from dataclasses import dataclass


@dataclass
class DataFile:
    """Represents a data file, e.g. a S3 object, or a local file."""

    file_path: str


class ScanTask(ABC):
    """Base class representing a unit of data to be read by a compute worker"""

    @abstractmethod
    def data_files(self) -> list[DataFile]:
        pass


@dataclass
class FileScanTask(ScanTask):
    """A unit of data in the form of data files"""

    data_file_list: list[DataFile]

    def data_files(self) -> list[DataFile]:
        return self.data_file_list


class ShardedScanTask(ScanTask):
    """A unit of data in the form of shards (e.g. shard 1-10 each represents 1/10 of all data in a Table)"""

    def data_files(self) -> list[DataFile]:
        raise NotImplementedError("data_files is not implemented for ShardedScanTask")



import logging
from typing import Dict, Optional, List, Tuple, Any
from deltacat import logs
from deltacat.constants import PYARROW_INFLATION_MULTIPLIER
from deltacat.compute.compactor_v2.constants import (
    AVERAGE_RECORD_SIZE_BYTES as DEFAULT_AVERAGE_RECORD_SIZE_BYTES,
)
from deltacat.compute.compactor_v2.model.merge_file_group import (
    LocalMergeFileGroupsProvider,
)
from deltacat.storage import (
    Manifest,
    ManifestEntry,
    metastore,
)
from deltacat.compute.compactor.model.delta_annotated import DeltaAnnotated
from deltacat.compute.compactor.model.round_completion_info import RoundCompletionInfo
from deltacat.compute.compactor_v2.utils.primary_key_index import (
    hash_group_index_to_hash_bucket_indices,
)
from deltacat.compute.resource_estimation.manifest import (
    estimate_manifest_entry_num_rows,
    estimate_manifest_entry_size_bytes,
    estimate_manifest_entry_column_size_bytes,
)
from deltacat.compute.resource_estimation.model import (
    EstimateResourcesParams,
    OperationType,
)
from deltacat.exceptions import RetryableError

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _get_task_options(
    cpu: float,
    memory: float,
    ray_custom_resources: Optional[Dict] = None,
    scheduling_strategy: str = "SPREAD",
) -> Dict:

    # NOTE: With DEFAULT scheduling strategy in Ray 2.20.0, autoscaler does
    # not spin up enough nodes fast and hence we see only approximately
    # 20 tasks get scheduled out of 100 tasks in queue. Hence, we use SPREAD
    # which is also ideal for merge and hash bucket tasks.
    # https://docs.ray.io/en/latest/ray-core/scheduling/index.html
    task_opts = {
        "num_cpus": cpu,
        "memory": memory,
        "scheduling_strategy": scheduling_strategy,
    }

    if ray_custom_resources:
        task_opts["resources"] = ray_custom_resources

    task_opts["max_retries"] = 3

    # List of possible botocore exceptions are available at
    # https://github.com/boto/botocore/blob/develop/botocore/exceptions.py
    task_opts["retry_exceptions"] = [RetryableError]

    return task_opts


def _get_merge_task_options(
    index: int,
    hb_group_idx: int,
    data_size: float,
    pk_size_bytes: float,
    num_rows: int,
    num_hash_groups: int,
    total_memory_buffer_percentage: int,
    incremental_index_array_size: int,
    debug_memory_params: Dict[str, Any],
    ray_custom_resources: Optional[Dict],
    estimate_resources_params: EstimateResourcesParams,
    round_completion_info: Optional[RoundCompletionInfo] = None,
    compacted_delta_manifest: Optional[Manifest] = None,
    primary_keys: Optional[List[str]] = None,
    memory_logs_enabled: Optional[bool] = None,
) -> Dict[str, Any]:
    if (
        round_completion_info
        and compacted_delta_manifest
        and round_completion_info.hb_index_to_entry_range
    ):
        logger.debug_conditional(
            f"[Merge task {index}]: Using previous compaction rounds to calculate merge memory: {round_completion_info.compacted_pyarrow_write_result}",
            memory_logs_enabled,
        )
        previous_inflation: float = (
            (
                round_completion_info.compacted_pyarrow_write_result.pyarrow_bytes
                / round_completion_info.compacted_pyarrow_write_result.file_bytes
            )
            if round_completion_info.compacted_pyarrow_write_result.file_bytes
            else PYARROW_INFLATION_MULTIPLIER
        )
        debug_memory_params["previous_inflation"] = previous_inflation

        average_record_size: float = (
            (
                round_completion_info.compacted_pyarrow_write_result.pyarrow_bytes
                / round_completion_info.compacted_pyarrow_write_result.records
            )
            if round_completion_info.compacted_pyarrow_write_result.records
            else DEFAULT_AVERAGE_RECORD_SIZE_BYTES
        )
        debug_memory_params["average_record_size"] = average_record_size

        iterable = hash_group_index_to_hash_bucket_indices(
            hb_group_idx, round_completion_info.hash_bucket_count, num_hash_groups
        )

        for hb_idx in iterable:
            if round_completion_info.hb_index_to_entry_range.get(str(hb_idx)) is None:
                continue

            entry_start, entry_end = round_completion_info.hb_index_to_entry_range[
                str(hb_idx)
            ]
            for entry_index in range(entry_start, entry_end):
                entry: ManifestEntry = compacted_delta_manifest.entries[entry_index]
                current_entry_size: float = (
                    estimate_manifest_entry_size_bytes(
                        entry=entry,
                        operation_type=OperationType.PYARROW_DOWNLOAD,
                        estimate_resources_params=estimate_resources_params,
                    )
                    or 0.0
                )
                current_entry_rows: int = (
                    estimate_manifest_entry_num_rows(
                        entry=entry,
                        operation_type=OperationType.PYARROW_DOWNLOAD,
                        estimate_resources_params=estimate_resources_params,
                    )
                    or 0
                )
                # NOTE: We can treat the current_entry_size and current_entry_rows as 0 as a None estimated entry size implies a 0 value
                data_size += current_entry_size
                num_rows += current_entry_rows
                if primary_keys:
                    pk_size: Optional[
                        float
                    ] = estimate_manifest_entry_column_size_bytes(
                        entry=entry,
                        columns=primary_keys,
                        operation_type=OperationType.PYARROW_DOWNLOAD,
                        estimate_resources_params=estimate_resources_params,
                    )
                    if not pk_size:
                        pk_size_bytes += current_entry_size
                    else:
                        pk_size_bytes += pk_size

    # total data downloaded + primary key hash column + pyarrow-to-numpy conversion
    # + primary key column + hashlib inefficiency + dict size for merge + incremental index array size
    total_memory = (
        data_size
        + pk_size_bytes
        + pk_size_bytes
        + num_rows * 20
        + num_rows * 20
        + num_rows * 20
        + incremental_index_array_size
    )
    debug_memory_params["data_size"] = data_size
    debug_memory_params["num_rows"] = num_rows
    debug_memory_params["pk_size_bytes"] = pk_size_bytes
    debug_memory_params["incremental_index_array_size"] = incremental_index_array_size
    debug_memory_params["total_memory"] = total_memory
    debug_memory_params["estimate_resources_params"] = estimate_resources_params

    total_memory = total_memory * (1 + total_memory_buffer_percentage / 100.0)
    debug_memory_params["total_memory_with_buffer"] = total_memory
    logger.debug_conditional(
        f"[Merge task {index}]: Params used for calculating merge memory: {debug_memory_params}",
        memory_logs_enabled,
    )
    return _get_task_options(0.01, total_memory, ray_custom_resources)


def hash_bucket_resource_options_provider(
    index: int,
    item: DeltaAnnotated,
    previous_inflation: float,
    average_record_size_bytes: float,
    total_memory_buffer_percentage: int,
    estimate_resources_params: EstimateResourcesParams,
    primary_keys: List[str] = None,
    ray_custom_resources: Optional[Dict] = None,
    memory_logs_enabled: Optional[bool] = None,
    **kwargs,
) -> Dict:
    debug_memory_params = {"hash_bucket_task_index": index}
    size_bytes = 0.0
    num_rows = 0
    total_pk_size = 0
    if not item.manifest or not item.manifest.entries:
        logger.debug(
            f"[Hash bucket task {index}]: No manifest entries, skipping memory allocation calculation"
        )
        return {"CPU": 0.01}

    for entry in item.manifest.entries:
        entry_size = estimate_manifest_entry_size_bytes(
            entry=entry,
            operation_type=OperationType.PYARROW_DOWNLOAD,
            estimate_resources_params=estimate_resources_params,
            **kwargs,
        )
        num_rows += estimate_manifest_entry_num_rows(
            entry=entry,
            operation_type=OperationType.PYARROW_DOWNLOAD,
            estimate_resources_params=estimate_resources_params,
            **kwargs,
        )
        size_bytes += entry_size

        if primary_keys:
            pk_size = estimate_manifest_entry_column_size_bytes(
                entry=entry,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                columns=primary_keys,
                estimate_resources_params=estimate_resources_params,
            )

            if pk_size is None:
                total_pk_size += entry_size
            else:
                total_pk_size += pk_size

    # total size + pk size + pyarrow-to-numpy conversion + pk hash column + hashlib inefficiency + hash bucket index column
    # Refer to hash_bucket step for more details.
    total_memory = (
        size_bytes
        + total_pk_size
        + total_pk_size
        + num_rows * 20
        + num_rows * 20
        + num_rows * 4
    )
    debug_memory_params["size_bytes"] = size_bytes
    debug_memory_params["num_rows"] = num_rows
    debug_memory_params["total_pk_size"] = total_pk_size
    debug_memory_params["total_memory"] = total_memory
    debug_memory_params["estimate_resources_params"] = estimate_resources_params

    debug_memory_params["previous_inflation"] = previous_inflation
    debug_memory_params["average_record_size_bytes"] = average_record_size_bytes

    # Consider buffer
    total_memory = total_memory * (1 + total_memory_buffer_percentage / 100.0)
    debug_memory_params["total_memory_with_buffer"] = total_memory
    logger.debug_conditional(
        f"[Hash bucket task {index}]: Params used for calculating hash bucketing memory: {debug_memory_params}",
        memory_logs_enabled,
    )

    return _get_task_options(0.01, total_memory, ray_custom_resources)


def merge_resource_options_provider(
    index: int,
    item: Tuple[int, List],
    num_hash_groups: int,
    hash_group_size_bytes: Dict[int, int],
    hash_group_num_rows: Dict[int, int],
    total_memory_buffer_percentage: int,
    estimate_resources_params: EstimateResourcesParams,
    round_completion_info: Optional[RoundCompletionInfo] = None,
    compacted_delta_manifest: Optional[Manifest] = None,
    ray_custom_resources: Optional[Dict] = None,
    primary_keys: Optional[List[str]] = None,
    memory_logs_enabled: Optional[bool] = None,
    **kwargs,
) -> Dict:
    debug_memory_params = {"merge_task_index": index}
    hb_group_idx = item[0]

    data_size = hash_group_size_bytes.get(hb_group_idx, 0)
    num_rows = hash_group_num_rows.get(hb_group_idx, 0)
    debug_memory_params["data_size_from_hash_group"] = data_size
    debug_memory_params["num_rows_from_hash_group"] = num_rows

    # upper bound for pk size of incremental
    pk_size_bytes = data_size
    incremental_index_array_size = num_rows * 4

    return _get_merge_task_options(
        index,
        hb_group_idx,
        data_size,
        pk_size_bytes,
        num_rows,
        num_hash_groups,
        total_memory_buffer_percentage,
        incremental_index_array_size,
        debug_memory_params,
        ray_custom_resources,
        round_completion_info=round_completion_info,
        compacted_delta_manifest=compacted_delta_manifest,
        primary_keys=primary_keys,
        memory_logs_enabled=memory_logs_enabled,
        estimate_resources_params=estimate_resources_params,
    )


def local_merge_resource_options_provider(
    estimated_da_size: float,
    estimated_num_rows: int,
    total_memory_buffer_percentage: int,
    estimate_resources_params: EstimateResourcesParams,
    round_completion_info: Optional[RoundCompletionInfo] = None,
    compacted_delta_manifest: Optional[Manifest] = None,
    ray_custom_resources: Optional[Dict] = None,
    primary_keys: Optional[List[str]] = None,
    deltacat_storage=metastore,
    deltacat_storage_kwargs: Optional[Dict] = {},
    memory_logs_enabled: Optional[bool] = None,
    **kwargs,
) -> Dict:
    index = hb_group_idx = LocalMergeFileGroupsProvider.LOCAL_HASH_BUCKET_INDEX
    debug_memory_params = {"merge_task_index": index}

    # upper bound for pk size of incremental
    pk_size_bytes = estimated_da_size
    incremental_index_array_size = estimated_num_rows * 4

    return _get_merge_task_options(
        index=index,
        hb_group_idx=hb_group_idx,
        data_size=estimated_da_size,
        pk_size_bytes=pk_size_bytes,
        num_rows=estimated_num_rows,
        num_hash_groups=1,
        incremental_index_array_size=incremental_index_array_size,
        total_memory_buffer_percentage=total_memory_buffer_percentage,
        debug_memory_params=debug_memory_params,
        ray_custom_resources=ray_custom_resources,
        round_completion_info=round_completion_info,
        compacted_delta_manifest=compacted_delta_manifest,
        primary_keys=primary_keys,
        memory_logs_enabled=memory_logs_enabled,
        estimate_resources_params=estimate_resources_params,
    )



import pytest

from deltacat.storage import (
    PartitionKey,
    PartitionScheme,
    IdentityTransform,
)


def test_partition_scheme_validates_empty_keys():
    # When creating a partition scheme with empty keys list
    with pytest.raises(
        ValueError, match="Partition scheme cannot have empty keys list"
    ):
        PartitionScheme.of(
            keys=[],
            name="test_partition_scheme",
            scheme_id="test_partition_scheme_id",
        )


def test_partition_scheme_validates_duplicate_keys():
    # When creating a partition scheme with duplicate keys
    with pytest.raises(ValueError, match="Duplicate partition key found: col1"):
        PartitionScheme.of(
            keys=[
                PartitionKey.of(
                    key=["col1"],
                    transform=IdentityTransform.of(),
                ),
                PartitionKey.of(
                    key=["col1"],  # Duplicate key
                    transform=IdentityTransform.of(),
                ),
            ],
            name="test_partition_scheme",
            scheme_id="test_partition_scheme_id",
        )


def test_partition_scheme_validates_duplicate_names():
    # When creating a partition scheme with duplicate partition key names
    with pytest.raises(
        ValueError, match="Duplicate partition key name found: partition_1"
    ):
        PartitionScheme.of(
            keys=[
                PartitionKey.of(
                    key=["col1"],
                    name="partition_1",
                    transform=IdentityTransform.of(),
                ),
                PartitionKey.of(
                    key=["col2"],  # Different field locator
                    name="partition_1",  # But duplicate name
                    transform=IdentityTransform.of(),
                ),
            ],
            name="test_partition_scheme",
            scheme_id="test_partition_scheme_id",
        )


def test_partition_scheme_allows_valid_keys():
    # When creating a partition scheme with valid keys
    partition_scheme = PartitionScheme.of(
        keys=[
            PartitionKey.of(
                key=["col1"],
                transform=IdentityTransform.of(),
            ),
            PartitionKey.of(
                key=["col2"],
                transform=IdentityTransform.of(),
            ),
        ],
        name="test_partition_scheme",
        scheme_id="test_partition_scheme_id",
    )

    # Then it should succeed
    assert partition_scheme is not None
    assert len(partition_scheme.keys) == 2
    assert partition_scheme.name == "test_partition_scheme"
    assert partition_scheme.id == "test_partition_scheme_id"



from typing import List, Dict
from collections import defaultdict
import uuid
from pyiceberg.table import Table
from pyiceberg.table.metadata import TableMetadata
from pyiceberg.table.snapshots import (
    Operation,
)
from pyiceberg.manifest import (
    DataFile,
    DataFileContent,
    ManifestContent,
    ManifestEntry,
    ManifestEntryStatus,
    ManifestFile,
    write_manifest,
)
import itertools
from pyiceberg.utils.concurrent import ExecutorFactory
from pyiceberg.table.update.snapshot import _SnapshotProducer, UpdateSnapshot


def replace_delete_files_override(
    update_snapshot: UpdateSnapshot,
) -> "_ReplaceDeleteFilesOverride":
    commit_uuid = uuid.uuid4()
    return _ReplaceDeleteFilesOverride(
        commit_uuid=commit_uuid,
        operation=Operation.OVERWRITE,
        transaction=update_snapshot._transaction,
        io=update_snapshot._io,
        snapshot_properties=update_snapshot._snapshot_properties,
    )


class _ReplaceDeleteFilesOverride(_SnapshotProducer):
    def _manifests(self) -> List[ManifestFile]:
        def _write_added_manifest() -> List[ManifestFile]:
            if self._added_data_files:
                with write_manifest(
                    format_version=self._transaction.table_metadata.format_version,
                    spec=self._transaction.table_metadata.spec(),
                    schema=self._transaction.table_metadata.schema(),
                    output_file=self.new_manifest_output(),
                    snapshot_id=self._snapshot_id,
                ) as writer:
                    for data_file in self._added_data_files:
                        writer.add(
                            ManifestEntry(
                                status=ManifestEntryStatus.ADDED,
                                snapshot_id=self._snapshot_id,
                                sequence_number=None,
                                file_sequence_number=None,
                                data_file=data_file,
                            )
                        )
                        writer.content = self.writer_content
                return [writer.to_manifest_file()]
            else:
                return []

        def _write_delete_manifest() -> List[ManifestFile]:
            # Check if we need to mark the files as deleted
            deleted_entries = self._deleted_entries()
            if len(deleted_entries) > 0:
                deleted_manifests = []
                partition_groups: Dict[int, List[ManifestEntry]] = defaultdict(list)
                for deleted_entry in deleted_entries:
                    partition_groups[deleted_entry.data_file.spec_id].append(
                        deleted_entry
                    )
                for spec_id, entries in partition_groups.items():
                    with write_manifest(
                        format_version=self._transaction.table_metadata.format_version,
                        spec=self._transaction.table_metadata.specs()[spec_id],
                        schema=self._transaction.table_metadata.schema(),
                        output_file=self.new_manifest_output(),
                        snapshot_id=self._snapshot_id,
                    ) as writer:
                        for entry in entries:
                            writer.add_entry(entry)
                    deleted_manifests.append(writer.to_manifest_file())
                return deleted_manifests
            else:
                return []

        executor = ExecutorFactory.get_or_create()

        added_manifests = executor.submit(_write_added_manifest)
        existing_manifests = executor.submit(self._existing_manifests)
        delete_manifests = executor.submit(_write_delete_manifest)
        return self._process_manifests(
            added_manifests.result()
            + existing_manifests.result()
            + delete_manifests.result()
        )

    def writer_content(self) -> ManifestContent:
        return ManifestContent.DELETES

    def _existing_manifests(self) -> List[ManifestFile]:
        """To determine if there are any existing manifest files.

        A fast append will add another ManifestFile to the ManifestList.
        All the existing manifest files are considered existing.
        """
        existing_manifests = []

        if self._parent_snapshot_id is not None:
            previous_snapshot = self._transaction.table_metadata.snapshot_by_id(
                self._parent_snapshot_id
            )

            if previous_snapshot is None:
                raise ValueError(
                    f"Snapshot could not be found: {self._parent_snapshot_id}"
                )

            for manifest in previous_snapshot.manifests(io=self._io):
                if (
                    manifest.has_added_files()
                    or manifest.has_existing_files()
                    or manifest.added_snapshot_id == self._snapshot_id
                ):
                    existing_manifests.append(manifest)

        return existing_manifests

    def _deleted_entries(self) -> List[ManifestEntry]:
        if self._parent_snapshot_id is not None:
            previous_snapshot = self._transaction.table_metadata.snapshot_by_id(
                self._parent_snapshot_id
            )
            if previous_snapshot is None:
                # This should never happen since you cannot overwrite an empty table
                raise ValueError(
                    f"Could not find the previous snapshot: {self._parent_snapshot_id}"
                )

            executor = ExecutorFactory.get_or_create()

            def _get_entries(manifest: ManifestFile) -> List[ManifestEntry]:
                return [
                    ManifestEntry(
                        status=ManifestEntryStatus.DELETED,
                        snapshot_id=entry.snapshot_id,
                        sequence_number=entry.sequence_number,
                        file_sequence_number=entry.file_sequence_number,
                        data_file=entry.data_file,
                    )
                    for entry in manifest.fetch_manifest_entry(
                        self._io, discard_deleted=True
                    )
                    if entry.data_file.content == DataFileContent.EQUALITY_DELETES
                    and entry.data_file in self._deleted_data_files
                ]

            list_of_entries = executor.map(
                _get_entries, previous_snapshot.manifests(self._io)
            )
            return list(itertools.chain(*list_of_entries))
        else:
            return []


def commit_append_snapshot(
    iceberg_table: Table, new_position_delete_files: List[DataFile]
) -> TableMetadata:
    tx = iceberg_table.transaction()
    try:
        if iceberg_table.metadata.name_mapping() is None:
            tx.set_properties(
                **{
                    "schema.name-mapping.default": tx.table_metadata.schema().name_mapping.model_dump_json()
                }
            )
        with append_delete_files_override(tx.update_snapshot()) as append_snapshot:
            if new_position_delete_files:
                for data_file in new_position_delete_files:
                    append_snapshot.append_data_file(data_file)
    except Exception as e:
        raise e
    else:
        return tx.commit_transaction().metadata


def append_delete_files_override(
    update_snapshot: UpdateSnapshot,
) -> "_AppendDeleteFilesOverride":
    commit_uuid = uuid.uuid4()
    return _AppendDeleteFilesOverride(
        commit_uuid=commit_uuid,
        operation=Operation.APPEND,
        transaction=update_snapshot._transaction,
        io=update_snapshot._io,
        snapshot_properties=update_snapshot._snapshot_properties,
    )


class _AppendDeleteFilesOverride(_SnapshotProducer):
    def _manifests(self) -> List[ManifestFile]:
        def _write_added_manifest() -> List[ManifestFile]:
            if self._added_data_files:
                with write_manifest(
                    format_version=self._transaction.table_metadata.format_version,
                    spec=self._transaction.table_metadata.spec(),
                    schema=self._transaction.table_metadata.schema(),
                    output_file=self.new_manifest_output(),
                    snapshot_id=self._snapshot_id,
                ) as writer:
                    for data_file in self._added_data_files:
                        writer.add(
                            ManifestEntry(
                                status=ManifestEntryStatus.ADDED,
                                snapshot_id=self._snapshot_id,
                                sequence_number=None,
                                file_sequence_number=None,
                                data_file=data_file,
                            )
                        )
                        writer.content = self.writer_content
                return [writer.to_manifest_file()]
            else:
                return []

        executor = ExecutorFactory.get_or_create()

        added_manifests = executor.submit(_write_added_manifest)
        existing_manifests = executor.submit(self._existing_manifests)

        return self._process_manifests(
            added_manifests.result() + existing_manifests.result()
        )

    def writer_content(self) -> ManifestContent:
        return ManifestContent.DELETES

    def _existing_manifests(self) -> List[ManifestFile]:
        """To determine if there are any existing manifest files.

        A fast append will add another ManifestFile to the ManifestList.
        All the existing manifest files are considered existing.
        """
        existing_manifests = []

        if self._parent_snapshot_id is not None:
            previous_snapshot = self._transaction.table_metadata.snapshot_by_id(
                self._parent_snapshot_id
            )

            if previous_snapshot is None:
                raise ValueError(
                    f"Snapshot could not be found: {self._parent_snapshot_id}"
                )

            for manifest in previous_snapshot.manifests(io=self._io):
                if (
                    manifest.has_added_files()
                    or manifest.has_existing_files()
                    or manifest.added_snapshot_id == self._snapshot_id
                ):
                    existing_manifests.append(manifest)

        return existing_manifests

    def _deleted_entries(self) -> List[ManifestEntry]:
        """To determine if we need to record any deleted manifest entries.

        In case of an append, nothing is deleted.
        """
        return []


def commit_replace_snapshot(
    iceberg_table: Table,
    new_position_delete_files: List[DataFile],
    to_be_deleted_files: List[DataFile],
) -> TableMetadata:
    tx = iceberg_table.transaction()
    try:
        if iceberg_table.metadata.name_mapping() is None:
            tx.set_properties(
                **{
                    "schema.name-mapping.default": tx.table_metadata.schema().name_mapping.model_dump_json()
                }
            )
        with replace_delete_files_override(
            tx.update_snapshot()
        ) as replace_delete_snapshot:
            if new_position_delete_files:
                for data_file in new_position_delete_files:
                    replace_delete_snapshot.append_data_file(data_file)
            if to_be_deleted_files:
                for delete_file in to_be_deleted_files:
                    replace_delete_snapshot.delete_data_file(delete_file)
    except Exception as e:
        raise e
    else:
        return tx.commit_transaction().metadata



from __future__ import annotations

from ray.types import ObjectRef

from typing import Any, Union

from abc import ABC, abstractmethod
from deltacat.io.ray_plasma_object_store import RayPlasmaObjectStore
from deltacat.storage import (
    LocalTable,
)
from deltacat.io.object_store import IObjectStore

LocalTableReference = Union[ObjectRef, LocalTable]


class LocalTableStorageStrategy(ABC):
    @property
    @abstractmethod
    def object_store(cls) -> IObjectStore:
        pass

    @abstractmethod
    def store_table(self, table: LocalTable) -> LocalTableReference:
        pass

    @abstractmethod
    def get_table(self, table_like: LocalTableReference) -> LocalTable:
        pass


class LocalTableRayObjectStoreReferenceStorageStrategy(LocalTableStorageStrategy):
    """
    Stores the table in the RayPlasmaObjectStore - see deltacat/io/ray_plasma_object_store.py
    """

    _object_store: IObjectStore = RayPlasmaObjectStore()

    @property
    def object_store(cls) -> IObjectStore:
        return cls._object_store

    def store_table(self, table: LocalTable) -> LocalTableReference:
        obj_ref: ObjectRef = self.object_store.put(table)
        return obj_ref

    def get_table(self, table_like: LocalTableReference) -> LocalTable:
        table = self.object_store.get(table_like)
        return table

    def get_table_reference(self, table_ref: Any) -> LocalTableReference:
        return self.object_store.deserialize_references([table_ref])[0]



from typing import List, Optional, Union
import pyarrow as pa
from deltacat.storage import Delta, Partition, PartitionLocator, DeltaLocator
from deltacat.storage import metastore
from deltacat.types.media import StorageType, ContentType
from deltacat.storage.model.schema import Schema


def create_delta_from_csv_file(
    namespace: str,
    file_paths: List[str],
    table_name: Optional[str] = None,
    table_version: int = 1,
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs,
) -> Delta:
    assert file_paths is not None, "file_paths cannot be empty"
    pa_table = create_table_from_csv_file_paths(file_paths)
    schema = Schema.of(pa_table.schema)
    staged_partition = stage_partition_from_file_paths(
        namespace,
        file_paths,
        schema,
        *args,
        table_name=table_name,
        table_version=table_version,
        **kwargs,
    )
    committed_delta = commit_delta_to_staged_partition(
        staged_partition,
        pa_table,
        content_type,
        *args,
        **kwargs,
    )
    return committed_delta


def create_table_from_csv_file_paths(
    file_paths: List[str],
) -> pa.Table:
    tables = []
    for file_path in file_paths:
        table = pa.csv.read_csv(file_path)
        tables.append(table)
    return pa.concat_tables(tables)


def stage_partition_from_file_paths(
    namespace: str,
    file_paths: List[str],
    schema: Schema,
    table_name: Optional[str] = None,
    table_version: int = 1,
    *args,
    **kwargs,
) -> Partition:
    if not metastore.namespace_exists(namespace, **kwargs):
        metastore.create_namespace(namespace, **kwargs)
    if table_name is None:
        table_name = "-".join(file_paths).replace("/", "_")
    metastore.create_table_version(
        namespace,
        table_name,
        str(table_version),
        schema=schema,
        **kwargs,
    )
    stream = metastore.get_stream(
        namespace,
        table_name,
        str(table_version),
        **kwargs,
    )
    staged_partition = metastore.stage_partition(stream, **kwargs)
    return staged_partition


def commit_delta_to_staged_partition(
    staged_partition,
    pa_table: pa.Table,
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs,
) -> Delta:
    committed_delta = commit_delta_to_partition(
        staged_partition,
        pa_table,
        content_type,
        *args,
        **kwargs,
    )
    metastore.commit_partition(staged_partition, **kwargs)
    return committed_delta


def download_delta(delta_like: Union[Delta, DeltaLocator], *args, **kwargs) -> Delta:
    return pa.concat_tables(
        metastore.download_delta(
            delta_like,
            storage_type=StorageType.LOCAL,
            *args,
            **kwargs,
        )
    )


def commit_delta_to_partition(
    partition: Union[Partition, PartitionLocator],
    pa_table: pa.Table = None,
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs,
) -> Delta:

    if isinstance(partition, PartitionLocator):
        partition = metastore.get_partition(
            partition.stream_locator, partition.partition_values, *args, **kwargs
        )

    staged_delta = metastore.stage_delta(
        pa_table,
        partition,
        content_type=content_type,
        **kwargs,
    )

    return metastore.commit_delta(staged_delta, **kwargs)



import pyarrow as pa
import pyarrow.compute as pc
import deltacat.compute.converter.utils.iceberg_columns as sc
from deltacat.compute.converter.utils.io import (
    download_data_table_and_append_iceberg_columns,
)
from deltacat.compute.converter.utils.converter_session_utils import (
    sort_data_files_maintaining_order,
)
import logging
from deltacat import logs
from typing import List, Dict, Tuple, Optional, Any
from pyiceberg.manifest import DataFile

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def dedupe_data_files(
    data_file_to_dedupe: List[Tuple[int, DataFile]],
    identifier_columns: List[str],
    remaining_data_table_after_convert: Optional[pa.Table],
    merge_sort_column: str,
    s3_client_kwargs: Optional[Dict[str, Any]],
) -> Tuple[pa.Table, int, int]:
    data_file_table = []
    if remaining_data_table_after_convert:
        data_file_table.append(remaining_data_table_after_convert)

    data_file_to_dedupe = sort_data_files_maintaining_order(
        data_files=data_file_to_dedupe
    )
    downloaded_data_file_record_count = 0
    for file_tuple in data_file_to_dedupe:
        data_file = file_tuple[1]
        data_file_to_dedupe_table = download_data_table_and_append_iceberg_columns(
            file=data_file,
            columns_to_download=identifier_columns,
            additional_columns_to_append=[
                sc._FILE_PATH_COLUMN_NAME,
                sc._ORDERED_RECORD_IDX_COLUMN_NAME,
            ],
            s3_client_kwargs=s3_client_kwargs,
        )
        logger.info(
            f"Length of downloaded data file table: {len(data_file_to_dedupe_table)}"
        )
        downloaded_data_file_record_count += len(data_file_to_dedupe_table)
        data_file_table.append(data_file_to_dedupe_table)

    final_data_to_dedupe = pa.concat_tables(data_file_table)

    dedupe_input_record_count = downloaded_data_file_record_count
    if remaining_data_table_after_convert:
        dedupe_input_record_count += len(remaining_data_table_after_convert)
    assert len(final_data_to_dedupe) == dedupe_input_record_count, (
        f"Mismatch record count while performing table concat, Got {len(final_data_to_dedupe)} in final table, "
        f"while input table length is: {dedupe_input_record_count}"
    )

    logger.info(f"Length of pyarrow table to dedupe:{len(final_data_to_dedupe)}")

    record_idx_iterator = iter(range(len(final_data_to_dedupe)))

    # Append global record index to used as aggregate column
    final_data_to_dedupe = sc.append_global_record_idx_column(
        final_data_to_dedupe, record_idx_iterator
    )

    final_data_table_indices = final_data_to_dedupe.group_by(
        sc._IDENTIFIER_COLUMNS_HASH_COLUMN_NAME, use_threads=False
    ).aggregate([(sc._GLOBAL_RECORD_IDX_COLUMN_NAME, "max")])

    pos_delete_indices = pc.invert(
        pc.is_in(
            final_data_to_dedupe[sc._GLOBAL_RECORD_IDX_COLUMN_NAME],
            value_set=final_data_table_indices[
                f"{sc._GLOBAL_RECORD_IDX_COLUMN_NAME}_max"
            ],
        )
    )

    final_data_table_to_delete = final_data_to_dedupe.filter(pos_delete_indices)

    final_data_table_to_delete = final_data_table_to_delete.drop(
        [sc._IDENTIFIER_COLUMNS_HASH_COLUMN_NAME, sc._GLOBAL_RECORD_IDX_COLUMN_NAME]
    )
    logger.info(
        f"Deduped {len(final_data_table_to_delete)} Records based off identifier columns."
    )
    return (
        final_data_table_to_delete,
        len(final_data_to_dedupe),
        int(final_data_to_dedupe.nbytes),
    )



import heapq
import logging

from collections import defaultdict
from typing import (
    Generator,
    Dict,
    Set,
    Type,
    TypeVar,
    NamedTuple,
    Any,
    List,
    Generic,
    AbstractSet,
)

from deltacat.experimental.storage.rivulet.metastore.delta import DeltaContext
from deltacat.experimental.storage.rivulet.metastore.sst import SSTableRow
from deltacat.experimental.storage.rivulet.metastore.sst_interval_tree import (
    OrderedBlockGroups,
    BlockGroup,
    Block,
)
from deltacat.experimental.storage.rivulet.reader.data_reader import (
    RowAndKey,
    FileReader,
)
from deltacat.experimental.storage.rivulet.reader.dataset_metastore import (
    DatasetMetastore,
)
from deltacat.experimental.storage.rivulet.reader.pyarrow_data_reader import (
    ArrowDataReader,
)
from deltacat.experimental.storage.rivulet.reader.query_expression import (
    QueryExpression,
)
from deltacat.experimental.storage.rivulet.reader.reader_type_registrar import (
    FileReaderRegistrar,
)
from deltacat.experimental.storage.rivulet import Schema
from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

FILE_FORMAT = TypeVar("FILE_FORMAT")
MEMORY_FORMAT = TypeVar("MEMORY_FORMAT")


class FileReaderWithContext(NamedTuple):
    reader: FileReader[FILE_FORMAT]
    context: DeltaContext


class ZipperMergeHeapRecord(NamedTuple):
    """
    Named tuple for data structure we're putting into heap during zipper merge

    Note we override the equality/comparison operators to use key
        so that we can add these items to a heap by key
    """

    key: Any
    data: FILE_FORMAT
    reader: FileReaderWithContext

    def __lt__(self, other):
        return self.key < other.key

    def __le__(self, other):
        return self.key <= other.key

    def __gt__(self, other):
        return self.key > other.key

    def __ge__(self, other):
        return self.key >= other.key


class ZipperBlockScanExecutor(Generic[MEMORY_FORMAT]):
    """
    Class for managing a zipper scan across multiple field groups. This class is only ever called inside the higher level BlockScanner class

    It is factored into a dedicated class because of the complexity and state management of
     zipper merging
    """

    def __init__(
        self,
        result_schema: Schema,
        deserialize_to: Type[MEMORY_FORMAT],
        ordered_block_groups: OrderedBlockGroups,
        query: QueryExpression[Any],
        metastore: DatasetMetastore,
        file_readers: Dict[str, FileReader],
    ):

        self.result_schema = result_schema
        self.deserialize_to = deserialize_to
        self.ordered_block_groups = ordered_block_groups
        self.query = query
        self.metastore = metastore
        self.file_readers = file_readers
        """
        Keeps track of block file readers that are open, across block group boundaries. E.g., if Block Group 1 has
        blocks [1,2,3] and BlockGroup2 has blocks [2,3], we will start reading blocks [2,3] and need to re-use the
        open iterator while reading BlockGroup2
        """
        self._open_file_readers: Dict[SSTableRow, FileReaderWithContext] = {}

    def scan(self) -> Generator[MEMORY_FORMAT, None, None]:
        """
        Perform N-wise zipper across N field groups.
        Within each field group, there is a set of blocks which belong in this BlockGroup's key range

        As a simplified example, we may have:
        FieldGroup1: [BlockA, BlockB]
        FieldGroup2: [BlockC]
        BlockA: keys 1,3,9,10
        BlockB: keys 2,4,5,6,7,8
        BlockC: keys 1-10

        The algorithm to merge these looks like:
        1. Load each block in DataReader to get iterator over sorted keys
        2. Build a heap of records across blocks across field groups
        3. Pop record(s) from heap as long as they have equal keys. For up to N records, merge column wise
        4. Continue until all blocks are read OR the key range in query is exceeded
        """
        for block_group in self.ordered_block_groups.block_groups:

            logger.debug(f"Starting scan of block group {block_group}")

            # Set of all blocks that need to be read within this block group
            blocks: set[Block] = {
                block
                for block_set in block_group.field_group_to_blocks.values()
                for block in block_set
            }
            # Open all file readers, such that self._open_block_iterators has pointers to open readers
            self.__open_file_readers(blocks)
            record_heap: List[ZipperMergeHeapRecord] = []

            # Seed record heap with record from each iterator
            file_reader_context: FileReaderWithContext
            for block, file_reader_context in self._open_file_readers.items():
                self.__push_next_row_back_to_heap(
                    block_group, file_reader_context, record_heap
                )

            # For each zipper merged entry from heap traversal, delegate to deserializer
            for zipper_merged in self.__zipper_merge_sorted_records(
                record_heap, block_group
            ):
                records = [z.data for z in self._dedupe_records(zipper_merged)]
                # TODO (multi format support) we need to handle joining across data readers in the future
                # For now, assume all data readers MUST read to Arrow intermediate format
                for result in ArrowDataReader().join_deserialize_records(
                    records, self.deserialize_to, self.result_schema.get_merge_key()
                ):
                    yield result

    def _dedupe_records(
        self, records: List[ZipperMergeHeapRecord]
    ) -> List[ZipperMergeHeapRecord]:
        """Deduplicate records with the same key (as a sorted list of records).

        Deduplication chooses records based on the following rules of precedence

        1. Levels with lower numbers take precedence over levels with higher numbers (L0 is preferred over L1)
        2. Newer stream positions take precedence over older stream positions

        Undefined Behavior:

        - Duplicate records within files from the same manifest (either in the same ir across data files)

        TODO: allow for the definition of a 'dedupe' column to break ties.
        """
        sort_criteria = lambda x: (
            -x.reader.context.level,
            x.reader.context.stream_position,
        )

        grouped_by_sort_group: defaultdict[
            Schema, List[ZipperMergeHeapRecord]
        ] = defaultdict(list)
        for record in records:
            grouped_by_sort_group[record.reader.context.schema].append(record)
        deduped = [
            max(group, key=sort_criteria) for group in grouped_by_sort_group.values()
        ]
        # Sort one last time across schemas (in case there's overlapping fields)
        deduped.sort(key=sort_criteria)
        return deduped

    def __zipper_merge_sorted_records(
        self, record_heap: List[ZipperMergeHeapRecord], block_group: BlockGroup
    ) -> Generator[List[ZipperMergeHeapRecord], None, None]:
        """
        Continually pop from heap until heap empty OR block range exceeded. Generate "zipper merge" of records

        Algorithm is:
        (1) Pop lowest element from heap. Includes pointer to the iterator it came from.
                Push next largest element from that generator back onto heap
        (2) Buffer records of same key and peek/pop the heap as long as there is a key match
                For any record popped, push next largest element from generator back onto heap
        (3) Yield merged record by invoking Data Reader

        This solution maintains the following invariants:
        (1) the heap will have at most N records, where N=total blocks in BlockGroup
        (2) the heap has the N smallest records globally
        (3) any data that needs to be merged for a given key exists in the heap

        :param record_heap: seeded heap of ZipperMergeHeapRecords.
        :param block_group: block group being traversed
        :return: generator of merged records. Note this is a list not a set to not require hash support
        """
        if not record_heap:
            return

        # Keep iterating until heap is empty or key range is exceeded
        while record_heap:
            curr_heap_record = heapq.heappop(record_heap)
            curr_pk = curr_heap_record.key

            if not self.query.matches_query(curr_pk):
                continue

            # Sanity check - assert that key we are looking at is in block group's range
            if not block_group.key_in_range(curr_pk):
                raise RuntimeError(
                    f"Did not expect to find key {curr_pk} on zipper merge heap"
                    f"for block group {block_group}"
                )

            # Find all records to be merged by continuing to pop heap
            merged_by_pk = [curr_heap_record]
            # For the current record itself - push next row back to heap
            self.__push_next_row_back_to_heap(
                block_group, curr_heap_record.reader, record_heap
            )
            # For the rest of the heap elements - peek/pop as long as they equal key
            # Note that heap[0] is equivalent to peek operation
            while record_heap and record_heap[0][0] == curr_pk:
                merge_heap_record: ZipperMergeHeapRecord = heapq.heappop(record_heap)
                merged_by_pk.append(merge_heap_record)
                self.__push_next_row_back_to_heap(
                    block_group, merge_heap_record.reader, record_heap
                )
            yield merged_by_pk

    def __push_next_row_back_to_heap(
        self,
        block_group: BlockGroup,
        row_context: FileReaderWithContext,
        record_heap: List[ZipperMergeHeapRecord],
    ):
        """
        This is a helper function for __zipper_merge_sorted_records and for scan().

        Given a file reader, it will next() records until it finds the next record within the block group
        and current query. It then pushes that record onto the heap

        Sometimes we end up needing to seek into the middle of a block because the key range of a query starts
        in the middle of the block. For example, if the block has keys range [0,100],
        and the query is for keys [50-100], we need to seek to the first key in the block that is >= 50

        TODO better support for seeking within block (rather than O(N) iteration)
        """

        file_reader = row_context.reader
        while file_reader.peek() is not None and (
            block_group.key_below_range(file_reader.peek().key)
            or self.query.below_query_range(file_reader.peek().key)
        ):
            try:
                # call next() on file reader to throw out key which is below range of block group
                next(file_reader)
            except StopIteration:
                # If we have exhausted iterator, this just means no keys from this block actually match the query
                file_reader.close()
                # TODO how to remove file reader from _open_file_readers?

        if (
            file_reader.peek()
            and self.query.matches_query(file_reader.peek().key)
            and block_group.key_in_range(file_reader.peek().key)
        ):
            try:
                r: RowAndKey = next(file_reader)
                heapq.heappush(
                    record_heap,
                    ZipperMergeHeapRecord(r.key, r.row, row_context),
                )
            except StopIteration:
                # This means we have exhausted the open FileReader and should close it
                file_reader.__exit__()
                # TODO how to remove file reader from _open_file_readers?

    def __open_file_readers(self, blocks: AbstractSet[Block]):
        """
        This method should be called once per block group.
        It opens iterators across all blocks in the block group and stores them in a map
        Blocks may already be open, if they were also in previous block groups.
        """
        for block in blocks:
            sst_row: SSTableRow = block.row
            if sst_row not in self._open_file_readers:
                file_reader = FileReaderRegistrar.construct_reader_instance(
                    sst_row,
                    self.metastore.file_provider,
                    self.result_schema.get_merge_key(),
                    self.result_schema,
                    self.file_readers,
                )
                file_reader.__enter__()
                # TODO we need some way to compare the blocks. using serialized timestamp as proxy for now
                context = FileReaderWithContext(file_reader, block.context)
                self._open_file_readers[sst_row] = context


class BlockScanner:
    """
    BlockScanner is a low level internal class which performs IO on Block Groups

    Note that we expect a block scanner to be initialized PER QUERY because it will keep state about ongoing execution,
      e.g. open iterators across block groups

    TODO efficiency improvements like parallelizing scanning.
    TODO handle "partial schema" use case, in which the query schema is a subset of full schema
    TODO in the future we will probably want to cache blocks read across queries
    """

    def __init__(self, metastore: DatasetMetastore):
        # Persist initialized file readers
        self.metastore = metastore
        self.file_readers: Dict[str, FileReader] = {}

    def scan(
        self,
        schema: Schema,
        deserialize_to: Type[MEMORY_FORMAT],
        blocks: Set[SSTableRow],
        query: QueryExpression[Any](),
    ) -> Generator[MEMORY_FORMAT, None, None]:
        """
        Scan records given query and deserialize to desired memory output format
        Set of blocks can all be scanned and returned independently
        TODO handle "partial schema" use case, in which the query schema is a subset of full schema
        TODO parallelize scan with async io
        """
        data_reader = ArrowDataReader()
        for block in blocks:
            file_reader = FileReaderRegistrar.construct_reader_instance(
                block,
                self.metastore.file_provider,
                schema.get_merge_key(),
                schema,
                self.file_readers,
            )
            with file_reader:
                for generated_records in file_reader.__iter__():
                    # Check whether row matches key in query before deserializing
                    if query.key_range:
                        start, end = query.key_range
                        if generated_records.key < start or generated_records.key > end:
                            continue

                    # Otherwise, key predicate matched and yield deserialized row
                    for deserialized_row in data_reader.deserialize_records(
                        generated_records, deserialize_to
                    ):
                        yield deserialized_row

    def scan_with_zipper(
        self,
        schema: Schema,
        deserialize_to: Type[MEMORY_FORMAT],
        ordered_block_groups: OrderedBlockGroups,
        query: QueryExpression[Any](),
    ) -> Generator[MEMORY_FORMAT, None, None]:
        zipper_scan_executor = ZipperBlockScanExecutor(
            schema,
            deserialize_to,
            ordered_block_groups,
            query,
            self.metastore,
            self.file_readers,
        )
        return zipper_scan_executor.scan()



#!/usr/bin/env python3
"""
DeltaCAT Table Monitor Job. Automatically runs data converter sessions in response to table updates.
"""

import argparse
import hashlib
import json
import logging
import os
import time
from typing import List, Optional

import pyarrow.fs as pafs
import ray
import deltacat

from pyiceberg.catalog import load_catalog, CatalogType
from pyiceberg.exceptions import NoSuchTableError

from deltacat import job_client, local_job_client
from deltacat.constants import DEFAULT_NAMESPACE
from deltacat.compute.converter.converter_session import converter_session
from deltacat.compute.converter.model.converter_session_params import (
    ConverterSessionParams,
)
from deltacat.compute.jobs.client import DeltaCatJobClient
from deltacat.utils.filesystem import (
    resolve_path_and_filesystem,
    FilesystemType,
)
import deltacat.logs as logs

# Initialize DeltaCAT logger
logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def monitor_table(
    catalog_type: str,
    warehouse_path: str,
    catalog_uri: Optional[str],
    namespace: str,
    table_name: str,
    merge_keys: List[str],
    filesystem_type: FilesystemType = FilesystemType.LOCAL,
    monitor_interval: float = 5.0,
    max_converter_parallelism: int = 1,
    ray_inactivity_timeout: int = 10,
) -> None:
    """Monitor an Iceberg table for changes and run converter sessions when needed."""

    logger.info(
        f"Starting table monitor. Namespace: '{namespace}', Table: '{table_name}', "
        f"Warehouse: '{warehouse_path}', Catalog type: '{catalog_type}', "
        f"Catalog URI: '{catalog_uri or 'None'}', Merge keys: '{merge_keys}', "
        f"Filesystem type: '{filesystem_type}', Monitor interval: '{monitor_interval}s', "
        f"Max converter parallelism: '{max_converter_parallelism}', "
        f"Ray inactivity timeout: '{ray_inactivity_timeout}s'"
    )

    # Create PyIceberg catalog
    catalog = load_catalog(
        "monitor_catalog",
        type=catalog_type,
        warehouse=warehouse_path,
        uri=catalog_uri or None,
    )

    # Set up filesystem
    filesystem = FilesystemType.to_filesystem(filesystem_type)
    if filesystem_type == FilesystemType.UNKNOWN:
        normalized_warehouse_path, filesystem = resolve_path_and_filesystem(
            warehouse_path
        )
        warehouse_path = normalized_warehouse_path

    logger.info(f"Resolved filesystem: {type(filesystem).__name__}")
    logger.info(f"Normalized warehouse path: {warehouse_path}")

    # Parse table identifier
    if not namespace:
        namespace = DEFAULT_NAMESPACE
    table_identifier = f"{namespace}.{table_name}"
    logger.info(f"  - Parsed table - namespace: '{namespace}', table: '{table_name}'")

    last_snapshot_id = None
    start_time = time.time()
    last_write_time = start_time  # Track last time we saw table activity

    while True:
        # Sleep before starting the first iteration and all subsequent iterations
        logger.debug(f"Sleeping for {monitor_interval}s before next check...")
        time.sleep(monitor_interval)

        logger.info(f"Checking table {table_identifier} for updates...")

        # Try to load the table
        try:
            tbl = catalog.load_table(table_identifier)
            current_snapshot_id = tbl.metadata.current_snapshot_id
            if last_snapshot_id != current_snapshot_id:
                logger.info(
                    f"New table version detected - snapshot ID: {current_snapshot_id}"
                )
                logger.info(f"Table has {len(tbl.metadata.snapshots)} snapshots")
                logger.info(f"Table format version: {tbl.metadata.format_version}")

                # Update last activity time when we detect table changes
                last_write_time = time.time()

                # Always run deduplication when there are snapshots (duplicates can exist within a single snapshot)
                logger.info(
                    f"Table has data - triggering converter session to resolve any duplicates..."
                )

                # Run converter session
                try:
                    converter_params = ConverterSessionParams.of(
                        {
                            "catalog": catalog,
                            "iceberg_namespace": namespace,
                            "iceberg_table_name": table_name,
                            "iceberg_warehouse_bucket_name": warehouse_path,
                            "merge_keys": merge_keys,
                            "enforce_primary_key_uniqueness": True,
                            "task_max_parallelism": max_converter_parallelism,
                            "filesystem": filesystem,
                            "location_provider_prefix_override": None,
                        }
                    )

                    logger.debug(f"Converter Session Parameters: {converter_params}")

                    logger.info(f"Starting converter session...")
                    updated_metadata = converter_session(params=converter_params)
                    logger.info(f"Converter session completed successfully")
                    current_snapshot_id = updated_metadata.current_snapshot_id
                    logger.info(
                        f"Current snapshot ID updated to: {current_snapshot_id}"
                    )
                except Exception as e:
                    logger.error(f"Converter session failed: {e}")
                    logger.error(f"Exception traceback:", exc_info=True)
                last_snapshot_id = current_snapshot_id
            else:
                logger.debug(
                    f"No table changes detected (snapshot ID: {current_snapshot_id})"
                )
        except NoSuchTableError:
            logger.info(f"Table {table_identifier} does not exist yet - waiting...")
        except Exception as e:
            logger.error(f"Error in table monitor: {e}")

        # Check for Ray inactivity timeout
        current_time = time.time()
        inactivity_duration = current_time - last_write_time

        if inactivity_duration >= ray_inactivity_timeout:
            logger.info(
                f"Ray inactivity timeout reached ({inactivity_duration:.1f}s >= {ray_inactivity_timeout}s)"
            )
            logger.info(
                f"No table activity detected for {inactivity_duration:.1f} seconds, shutting down Ray..."
            )

            try:
                if ray.is_initialized():
                    ray.shutdown()
                    logger.info("Ray shutdown successfully due to inactivity")
                else:
                    logger.info("Ray was not initialized, nothing to shut down")
            except Exception as e:
                logger.error(f"Error shutting down Ray: {e}")

            logger.info(f"Table monitor stopping due to inactivity timeout")
            break

    logger.info(f"Table monitor completed")


def _generate_job_name(warehouse_path: str, namespace: str, table_name: str) -> str:
    """
    Generate a unique job name based on warehouse path, namespace, and table name.

    Args:
        warehouse_path: Warehouse path
        namespace: Table namespace
        table_name: Table name

    Returns:
        Job name string.
    """
    # Create a sha1 digest of the warehouse path, namespace, and table name
    digest = hashlib.sha1(
        f"{warehouse_path}-{namespace}-{table_name}".encode()
    ).hexdigest()
    job_name = f"deltacat-monitor-{digest}"

    return job_name


def _cleanup_terminated_jobs_for_submission_id(
    client: DeltaCatJobClient, submission_id: str
) -> bool:
    """Clean up any terminated jobs with the given submission ID."""
    logger.debug(
        f"Searching for terminated jobs to cleanup with submission ID: {submission_id}"
    )
    try:
        all_jobs = client.list_jobs()
        logger.debug(f"All jobs: {all_jobs}")
        for job in all_jobs:
            if job.submission_id == submission_id and job.status.is_terminal():
                logger.info(
                    f"Cleaning up terminated job: {submission_id} (status: {job.status})"
                )
                client.delete_job(submission_id)
                return True
    except Exception as e:
        logger.warning(f"Cleanup failed for job '{submission_id}': {e}")
    return False


def submit_table_monitor_job(
    warehouse_path: str,
    catalog_type: CatalogType,
    catalog_uri: Optional[str],
    namespace: str,
    table_name: str,
    merge_keys: list,
    monitor_interval: float,
    max_converter_parallelism: int,
    filesystem: pafs.FileSystem = None,
    cluster_cfg_file_path: Optional[str] = None,
    ray_inactivity_timeout: int = 10,
) -> str:
    """
    Submit a table monitor job to Ray cluster.

    Args:
        warehouse_path: Warehouse path
        catalog_type: Catalog type
        catalog_uri: Catalog URI
        namespace: Table namespace
        table_name: Table name to monitor
        merge_keys: List of merge key column names
        monitor_interval: Seconds between monitoring checks
        max_converter_parallelism: Maximum number of concurrent converter tasks
        filesystem: PyArrow filesystem instance
        cluster_cfg_file_path: Path to cluster config file (None for local)
        ray_inactivity_timeout: Seconds to wait before shutting down Ray cluster
    Returns:
        Job ID of the submitted job
    """

    # Parse table identifier to extract namespace and table name
    if not namespace:
        namespace = DEFAULT_NAMESPACE

    # Generate unique job ID based on the warehouse and table path
    job_name = _generate_job_name(
        warehouse_path=warehouse_path, namespace=namespace, table_name=table_name
    )

    # Resolve the appropriate local or remote job client
    if cluster_cfg_file_path:
        # Submit to remote cluster
        logger.info(
            f"Preparing to submit job to remote cluster: {cluster_cfg_file_path}"
        )
        # Set the cluster name to the job ID to prevent starting multiple Ray clusters monitoring the same table.
        client = job_client(cluster_cfg_file_path, cluster_name_override=job_name)
    else:
        # Submit to local cluster using DeltaCAT local job client
        ray_init_args = {
            "local_mode": True,
            "resources": {"convert_task": max_converter_parallelism},
        }
        logger.info(
            f"Preparing to submit job locally with ray init args: {ray_init_args}"
        )
        client = local_job_client(ray_init_args=ray_init_args)

    # Add filesystem type - determine from filesystem instance
    filesystem_type = FilesystemType.from_filesystem(filesystem)

    # Build CLI arguments for table_monitor job
    table_monitor_script_dir = os.path.dirname(os.path.abspath(__file__))
    table_monitor_script_path = os.path.join(
        table_monitor_script_dir, "table_monitor.py"
    )

    logger.debug(f"Table monitor script path: {table_monitor_script_path}")
    logger.debug(
        f"Table monitor script exists: {os.path.exists(table_monitor_script_path)}"
    )

    cmd_args = [
        f"python {table_monitor_script_path}",
        f"--catalog-type '{catalog_type.value}'",
        f"--warehouse-path '{warehouse_path}'",
        f"--catalog-uri '{catalog_uri}'",
        f"--namespace '{namespace}'",
        f"--table-name '{table_name}'",
        f"--merge-keys '{json.dumps(merge_keys)}'",
        f"--monitor-interval {monitor_interval}",
        f"--max-converter-parallelism {max_converter_parallelism}",
        f"--ray-inactivity-timeout {ray_inactivity_timeout}",
        f"--filesystem-type '{filesystem_type}'",
    ]

    # Join all arguments
    entrypoint = " ".join(cmd_args)
    logger.debug(
        f"Submitting table monitor job '{job_name}' with entrypoint: {entrypoint}"
    )

    # Clean up any terminated jobs with the same submission ID to allow reuse
    _cleanup_terminated_jobs_for_submission_id(client, job_name)

    # Submit the job with the correct working directory
    # Working directory should be the converter_agent directory where table_monitor.py is located
    job_submission_id = client.submit_job(
        submission_id=job_name,
        entrypoint=entrypoint,
        runtime_env={"working_dir": table_monitor_script_dir},
    )

    logger.info(f"Table monitor job submitted successfully: {job_submission_id}")

    return job_submission_id


def run(
    catalog_type: str,
    warehouse_path: str,
    catalog_uri: Optional[str],
    namespace: str,
    table_name: str,
    merge_keys: str,
    filesystem_type: str = "local",
    monitor_interval: float = 1.0,
    max_converter_parallelism: int = 1,
    ray_inactivity_timeout: int = 10,
) -> None:
    """Run table monitor with the given parameters."""

    # Parse merge keys
    merge_keys_list = json.loads(merge_keys)

    # Run the monitor
    monitor_table(
        catalog_type=catalog_type,
        warehouse_path=warehouse_path,
        catalog_uri=catalog_uri,
        namespace=namespace,
        table_name=table_name,
        merge_keys=merge_keys_list,
        filesystem_type=filesystem_type,
        monitor_interval=monitor_interval,
        max_converter_parallelism=max_converter_parallelism,
        ray_inactivity_timeout=ray_inactivity_timeout,
    )


if __name__ == "__main__":
    """
    DeltaCAT Table Monitor - Monitor Iceberg tables and run converter sessions

    Example usage:
    $ python table_monitor.py \
    $   --catalog-type 'rest' \
    $   --warehouse-path '/tmp/iceberg-warehouse' \
    $   --catalog-uri 'http://localhost:8181' \
    $   --namespace 'default' \
    $   --table-name 'demo_table' \
    $   --merge-keys '["id"]' \
    $   --monitor-interval 1.0 \
    $   --max-converter-parallelism 2 \
    $   --ray-inactivity-timeout 300.0
    """

    script_args = [
        (
            ["--catalog-type"],
            {
                "help": "Catalog type name (rest, hive, sql)",
                "type": str,
                "required": True,
            },
        ),
        (
            ["--warehouse-path"],
            {
                "help": "Warehouse path",
                "type": str,
                "required": True,
            },
        ),
        (
            ["--catalog-uri"],
            {
                "help": "Catalog URI",
                "type": str,
                "required": True,
            },
        ),
        (
            ["--namespace"],
            {
                "help": "Table namespace",
                "type": str,
                "required": True,
            },
        ),
        (
            ["--table-name"],
            {
                "help": "Table name to monitor",
                "type": str,
                "required": True,
            },
        ),
        (
            ["--merge-keys"],
            {
                "help": "Comma-separated merge key column names",
                "type": str,
                "required": True,
            },
        ),
        (
            ["--filesystem-type"],
            {
                "help": "Filesystem type",
                "type": str,
                "default": "local",
            },
        ),
        (
            ["--monitor-interval"],
            {
                "help": "Seconds between monitoring checks",
                "type": float,
                "default": 5.0,
            },
        ),
        (
            ["--max-converter-parallelism"],
            {
                "help": "Maximum number of concurrent converter tasks",
                "type": int,
                "default": 1,
            },
        ),
        (
            ["--ray-inactivity-timeout"],
            {
                "help": "Ray inactivity timeout in seconds (Ray will shutdown if no activity)",
                "type": int,
                "default": 300,
            },
        ),
    ]

    # Parse CLI input arguments
    parser = argparse.ArgumentParser(
        description="DeltaCAT Table Monitor - Monitor Iceberg tables and run converter sessions"
    )
    for args, kwargs in script_args:
        parser.add_argument(*args, **kwargs)
    args = parser.parse_args()
    print(f"[TABLE MONITOR] Command Line Arguments: {args}")

    # Initialize DeltaCAT
    deltacat.init()

    # Run the table monitor using the parsed arguments
    run(**vars(args))



import ray
import deltacat
import daft


def print_package_version_info():
    print(f"DeltaCAT Version: {deltacat.__version__}")
    print(f"Ray Version: {ray.__version__}")
    print(f"Daft Version: {daft.__version__}")


@ray.remote
def hello_worker():
    print("Hello, Worker!")
    print_package_version_info()


def run():
    print("Hello, Driver!")
    print_package_version_info()
    hello_worker.remote()


if __name__ == "__main__":
    # initialize deltacat
    deltacat.init()

    # run the example
    run()



from typing import List, Optional, Union, Dict, Any

from ray.data import Dataset as RayDataset
from ray.data import read_datasource

from deltacat.io.datasource.deltacat_datasource import DeltaCatDatasource
from deltacat.io.dataset.deltacat_dataset import DeltaCatDataset
from deltacat.utils.common import ReadKwargsProvider
from deltacat.utils.url import DeltaCatUrl, DeltaCatUrlReader
from deltacat.io.datasource.deltacat_datasource import DeltacatReadType


class EmptyReadKwargsProvider(ReadKwargsProvider):
    def _get_kwargs(
        self,
        datasource_type: str,
        kwargs: Dict[str, Any],
    ) -> Dict[str, Any]:
        return {}


def read_deltacat(
    urls: Union[DeltaCatUrl, List[DeltaCatUrl]],
    *,
    deltacat_read_type: DeltacatReadType = DeltacatReadType.DATA,
    timestamp_as_of: Optional[int] = None,
    merge_on_read: Optional[bool] = False,
    read_kwargs_provider: Optional[ReadKwargsProvider] = EmptyReadKwargsProvider(),
) -> DeltaCatDataset:
    """Reads the given DeltaCAT URLs into a Ray Dataset. DeltaCAT URLs can
    either reference objects registered in a DeltaCAT catalog, or unregistered
    external objects that are readable into a Ray Dataset.

    Unless `metadata_only` is `True`, all reads of registered DeltaCAT catalog
    object data must resolve to a single table version.

    When reading unregistered external objects, all additional keyword
    arguments specified are passed into the Ray Datasource resolved for the
    given DeltaCAT URLs.

    Examples:
        >>> # Read the latest active DeltaCAT table version:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("dc://my_catalog/my_namespace/my_table")
        >>> # If `my_catalog is the default catalog, this is equivalent to:
        >>> dc.io.read_deltacat("namespace://my_namespace/my_table")
        >>> # If `my_namespace` is the default namespace, this is equivalent to:
        >>> dc.io.read_deltacat("table://my_table")

        >>> # Read metadata from all partitions and deltas of the latest active
        >>> # DeltaCAT table version:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("dc://my_catalog/my_namespace/my_table", metadata_only=True)
        >>> # Since "default" always resolves to the latest active table version.
        >>> # This is equivalent to:
        >>> dc.io.read_deltacat("dc://my_catalog/my_namespace/my_table/default", metadata_only=True)

        >>> # Read only the latest active table version's top-level metadata:
        >>> dc.io.read_deltacat("dc://my_catalog/my_namespace/my_table/default", metadata_only=True, recursive=False)

        >>> # Read only top-level metadata from a DeltaCAT table:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("dc://my_catalog/my_namespace/my_table", metadata_only=True, recursive=False)

        >>> # Read top-level table metadata from all table versions:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("dc://my_catalog/my_namespace/my_table/*", metadata_only=True, recursive=False)

        >>> # Read metadata from all partitions and deltas of all table versions:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("dc://my_catalog/my_namespace/my_table/*", metadata_only=True)

        >>> # Read metadata from all tables and table versions of the namespace:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("dc://my_catalog/my_namespace/*", metadata_only=True)

        >>> # Read metadata from the latest active table version for each
        >>> # table in the namespace:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("dc://my_catalog/my_namespace", metadata_only=True)

        >>> # Read metadata from the latest active table version for each
        >>> # table in the namespace:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("dc://my_catalog/my_namespace", metadata_only=True)

        >>> # Read metadata from the latest active table version for each
        >>> # table in the catalog's default namespace:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("dc://my_catalog", metadata_only=True)

        >>> # Read metadata from all table versions for each table in each
        >>> # catalog namespace:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("dc://my_catalog/*", metadata_only=True)

        >>> # Read the Iceberg stream of the latest active DeltaCAT table version,
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("dc://my_catalog/my_namespace/my_table/default/iceberg")
        >>> # Or, if `my_catalog` is the default catalog, this is equivalent to:
        >>> dc.io.read_deltacat("namespace://my_namespace/my_table/default/iceberg")
        >>> # Or, if `my_namespace` is the default namespace, this is equivalent to:
        >>> dc.io.read_deltacat("table://my_table/default/iceberg")

        >>> # Read an external unregistered Iceberg table `my_db.my_table`:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("iceberg://my_db.my_table")

        >>> # Read an external unregistered audio file from /my/audio.mp4:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("audio+file:///my/audio.mp4")

        >>> # Read an external unregistered audio file from s3://my/audio.mp4:
        >>> import deltacat as dc
        >>> dc.io.read_deltacat("audio+s3://my/audio.mp4")

    Args:
        urls: The DeltaCAT URLs to read.
        deltacat_read_type: If METADATA, reads only DeltaCAT metadata for the
            given URL and skips both recursive metadata expansion and reads
            of the underlying data files. If METADATA_RECURSIVE then recursively
            expands child metadata but does not read underlying data files. If
            DATA then recursively expands child metadata to discover and read
            all underlying data files.
        timestamp_as_of: Reads a historic snapshot of the given paths as-of the
            given millisecond-precision epoch timestamp (only used when reading
            registered DeltaCAT catalog objects).
        merge_on_read: If True, merges all unmaterialized inserts, updates,
            and deletes in the registered DeltaCAT table version being read. Only
            applicable if `metadata_only` is False.
            ray_remote_args: kwargs passed to `ray.remote` in the read tasks.
        read_kwargs_provider: Resolves
            :class:`~deltacat.types.media.DatasourceType` string keys to
            kwarg dictionaries to pass to the resolved
            :class:`~ray.data.Datasource` implementation for each distinct
            DeltaCAT URL type.

    Returns:
        DeltacatDataset holding Arrow records read from the specified URL.
    """
    # TODO(pdames): The below implementation serializes reads of each URL and
    #   then unions their respective datasets together. While this was an easy
    #   starting point to implement, a more efficient implementation should push
    #   all URLs down into `DeltacatDatasource` to parallelize all reads
    #   (i.e., by returning the `ReadTask` for all datasources in
    #   `get_read_tasks()` and estimating the corresponding memory size across
    #   all datasources in `estimate_inmemory_data_size()`.
    dataset: RayDataset = None
    for url in urls:
        if not url.is_deltacat_catalog_url():
            # this URL points to an external unregistered Ray Datasource
            # TODO(pdames): Honor metadata only reads of external datasources
            #  by registering only file paths & metadata in delta manifests.
            reader = DeltaCatUrlReader(url)
            next_ds = reader.read(read_kwargs_provider(url.datastore_type, {}))
        else:
            # this URL points to a registered DeltaCAT object
            next_ds = read_datasource(
                DeltaCatDatasource(
                    url=url,
                    deltacat_read_type=deltacat_read_type,
                    timestamp_as_of=timestamp_as_of,
                    merge_on_read=merge_on_read,
                    read_kwargs_provider=read_kwargs_provider,
                )
            )
        # union the last dataset read into the result set
        if not dataset:
            dataset = next_ds
        else:
            dataset.union(next_ds)
    return DeltaCatDataset.from_dataset(dataset)



from typing import Protocol, Iterable, List, Union, Any, Dict

from deltacat.experimental.storage.rivulet.metastore.sst import SSTableRow
import pyarrow as pa

MEMTABLE_DATA = Union[Iterable[Dict[str, Any]], pa.Table]


class DataSerializer(Protocol):
    """
    Interface for writing data only.

    As data is written, it must emit sufficient metadata to build SSTable
    Each format will have a specific data writer (e.g. ParquetDataWriter)

    TODO future improvements:
        1. How does data writer control how it chooses to write to existing files vs new files?
            For now, we will not expose this configuration and always write each batch to
            a new file
        2. Related to 1, how should we expose URI(s) to write to? Probably DataWriter can
            use FileProvider and needs to know relevant ids like task ID.
    """

    def flush_batch(self, sorted_records: MEMTABLE_DATA) -> List[SSTableRow]:
        """
        Flushes rows to file, and return appropriate metadata to build SSTable

        TODO future improvements
         1. Finalize type for input records (instead of MvpRow)

            Options could be:
            (a) Something like Iceberg "StructLike" which allows flexible integrations without memcopy for row-oriented formats, e.g. can make Spark InternalRow structlike
            (b) use arrow. We will probably use arrow for writing parquet, although
                probably it isn't ideal for row oriented formats
         2. Keep in mind, most implementation of DataWriter will be written in rust.

        :param sorted_records: Records sorted by key
        :return: metadata used to build SS Table
        """
        ...



import pyarrow as pa
from typing import Union, Iterator, Any
import numpy as np

# Refer to: https://iceberg.apache.org/spec/#reserved-field-ids for reserved field ids
ICEBERG_RESERVED_FIELD_ID_FOR_FILE_PATH_COLUMN = 2147483546

# Refer to: https://iceberg.apache.org/spec/#reserved-field-ids for reserved field ids
ICEBERG_RESERVED_FIELD_ID_FOR_POS_COLUMN = 2147483545


def _get_iceberg_col_name(suffix: str) -> str:
    return suffix


_ORDERED_RECORD_IDX_COLUMN_NAME = _get_iceberg_col_name("pos")
_ORDERED_RECORD_IDX_COLUMN_TYPE = pa.int64()
_ORDERED_RECORD_IDX_FIELD_METADATA = {
    b"PARQUET:field_id": f"{ICEBERG_RESERVED_FIELD_ID_FOR_POS_COLUMN}"
}
_ORDERED_RECORD_IDX_COLUMN_FIELD = pa.field(
    _ORDERED_RECORD_IDX_COLUMN_NAME,
    _ORDERED_RECORD_IDX_COLUMN_TYPE,
    metadata=_ORDERED_RECORD_IDX_FIELD_METADATA,
    nullable=False,
)


def get_record_index_column_array(obj: Any) -> Union[pa.Array, pa.ChunkedArray]:
    return pa.array(
        obj,
        _ORDERED_RECORD_IDX_COLUMN_TYPE,
    )


def append_record_idx_col(
    table: pa.Table, ordered_record_indices: Iterator[int]
) -> pa.Table:
    table = table.append_column(
        _ORDERED_RECORD_IDX_COLUMN_FIELD,
        get_record_index_column_array(ordered_record_indices),
    )
    return table


_FILE_PATH_COLUMN_NAME = _get_iceberg_col_name("file_path")
_FILE_PATH_COLUMN_TYPE = pa.string()
_FILE_PATH_FIELD_METADATA = {
    b"PARQUET:field_id": f"{ICEBERG_RESERVED_FIELD_ID_FOR_FILE_PATH_COLUMN}"
}
_FILE_PATH_COLUMN_FIELD = pa.field(
    _FILE_PATH_COLUMN_NAME,
    _FILE_PATH_COLUMN_TYPE,
    metadata=_FILE_PATH_FIELD_METADATA,
    nullable=False,
)


def append_file_path_column(table: pa.Table, file_path: str) -> pa.Table:
    table = table.append_column(
        _FILE_PATH_COLUMN_FIELD,
        pa.array(np.repeat(file_path, len(table)), _FILE_PATH_COLUMN_TYPE),
    )
    return table


_GLOBAL_RECORD_IDX_COLUMN_NAME = _get_iceberg_col_name("global_record_index")
_GLOBAL_RECORD_IDX_COLUMN_TYPE = pa.int64()
_GLOBAL_RECORD_IDX_COLUMN_FIELD = pa.field(
    _GLOBAL_RECORD_IDX_COLUMN_NAME,
    _GLOBAL_RECORD_IDX_COLUMN_TYPE,
)


def append_global_record_idx_column(
    table: pa.Table, ordered_record_indices: Iterator[int]
) -> pa.Table:
    table = table.append_column(
        _GLOBAL_RECORD_IDX_COLUMN_NAME,
        pa.array(ordered_record_indices, _GLOBAL_RECORD_IDX_COLUMN_TYPE),
    )
    return table


_IDENTIFIER_COLUMNS_HASH_COLUMN_NAME = _get_iceberg_col_name(
    "identifier_columns_hashed"
)
